#include <cstdlib>
#include <string>
#include <iostream>
#include "astrowav.h"
#include "ballyBinFile.h"
#include "Error.h"

using namespace std;

unsigned char binBuffer[0x6000];
unsigned char loadBuffer[0x1000];

void getParmString(int *argc, char ***argv, char *returnString, int maxLength)
{
	if (strlen((*argv)[0]) == 1)
	{
		if (*argc == 1)
			throw Error::badParmValue();

		if ((*argv)[1][0] == '-')
			throw Error::badParmValue();

		strncpy(returnString, (*argv)[1], maxLength);
		--(*argc);
		(*++(*argv));
	} else
	{
		strncpy(returnString, **argv + 1, maxLength);
	}
	(**argv)[1] = 0;
}

int main(int argc, char *argv[])
{
	char binName[256]   = "",
         wavName[256]   = "",
	     tapeName[256]  = "",
	     parmString[20] = "";

	int  binByteNum   = 0,
         loadByteNum  = 0,
         channelCount = 2,
         bitCount     = 8,
         c            = 0,
	     verbosity    = 1;

	long sampleRate = 22050;

	double pHeaderLength = -1.0,
		   pFooterLength = -1.0,
		   pGapLength    = -1.0;

	ofstream ofsScBin;
	BallyBinFile inFile;

	if (argc == 1)
	{
		cout << "AstroWav v2.1\n";
		cout << "Released January 10, 2004\n";
		cout << "Written by: Bob Colbert\n\n";
		cout << "Usage:\n";
		cout << "   AstroWav [options] (input filename - names with spaces require quotes)\n";
		cout << "\n";
		cout << "Options:\n";
		cout << "   [-b(bits)]      Number of bits per sample in .wav file: (8) or 16.\n";
		cout << "   [-c(channels)]  Number of audio channels: (1) or 2.\n";
		cout << "   [-f(length)]    Footer length in seconds, decimals allowed.\n";
		cout << "                   If no value is provided, the original footer length is used.\n";
		cout << "   [-g(length)]    Gap length in seconds, decimals allowed.\n";
		cout << "                   If no value is provided, the original gap length is used.\n";
		cout << "   [-h(length)]    Header length in seconds, decimals allowed.\n";
		cout << "                   If no value is provided, the original header length is used.\n";
		cout << "   [-k(rate)]      Sample rate: (22050) or 44100.\n";
		cout << "   [-s(name)]      Name of bitmap file to use as the load screen.\n";
        cout << "   [-v(verbosity)] Verbosity: 0,(1) or 2.  0 is silent and 2 is the most detailed.\n";
		cout << "   [-w(name)]      Name of the output .wav file.  Names with spaces require quotes.\n\n";

		return 1;
	}

	while (--argc > 0 && (*++argv)[0] == '-')
    {
		while (c = *++argv[0])
        {
			switch (c)
			{
				case 'b':
					getParmString(&argc, &argv, parmString, 19);
					bitCount = atoi(parmString);
					break;
				case 'c':
					getParmString(&argc, &argv, parmString, 19);
					channelCount = atoi(parmString);
					break;
				case 'f':
					getParmString(&argc, &argv, parmString, 19);
					pFooterLength = atof(parmString);
					break;
				case 'g':
					getParmString(&argc, &argv, parmString, 19);
					pGapLength = atof(parmString);
					break;
				case 'h':
					getParmString(&argc, &argv, parmString, 19);
					pHeaderLength = atof(parmString);
					break;
				case 'k':
					getParmString(&argc, &argv, parmString, 19);
					sampleRate = atoi(parmString);
					break;
				case 's':
					getParmString(&argc, &argv, parmString, 255);
					inFile.setScreenBmpName(parmString);
					break;
                case 'v':
					getParmString(&argc, &argv, parmString, 19);
                    verbosity = atoi(parmString);
                    break;
				case 'w':
					getParmString(&argc, &argv, wavName, 255);
					break;
				default:
					cerr << "Invalid option " << (char)c << "\n";
					return 1;
					break;
			}
        }
    }

	if (verbosity)
	{
		cout << "AstroWav v2.0beta\n";
		cout << "Released January 31, 2004\n";
		cout << "Written by: Bob Colbert\n\n";
		cout.flush();
	}

    try
    {
        inFile.open(argv[0],"r",verbosity);
        if (!strlen(wavName))
        {
			BallyBinFile::cleanFileName(argv[0], ".bin");
			BallyBinFile::cleanFileName(argv[0], ".bml");
            sprintf(wavName, "%s.wav", argv[0]);
        }
		inFile.pHeaderLength = pHeaderLength;
		inFile.pFooterLength = pFooterLength;
		inFile.pGapLength = pGapLength;
        inFile.writeWavFile(wavName, sampleRate, bitCount, channelCount);
		if (verbosity)
        {
			cout << "Successfully created .wav file "<< wavName<<"!\n";
        }
    }

    catch(Error::endOfFile)
    {
        cerr << "Unexpected end of input reached\n";
        return 1;
    }

	catch(Error::badOutputFileName)
	{
		cerr << "Invalid output .wav filename\n";
		return 1;
	}

	return 0;
}
