// BallyWavIn.cpp: implementation of the BallyWavIn class.
//
//////////////////////////////////////////////////////////////////////

#ifdef _AFXDLL
    #include "stdafx.h"
#endif
#include "BallyWavIn.h"
#include "Error.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

BallyWavIn::BallyWavIn()
{

}

BallyWavIn::~BallyWavIn()
{

}

int BallyWavIn::open(char *filename, int sign, int pMinHalfWavLength, int pVerbosity)
{
    int returnVal =
        WavDataIn::open(filename, sign, pMinHalfWavLength, pVerbosity);

    try
    {
	    firstBit = 1;
	    syncWav();
    }

    catch (Error::endOfFile)
    {
        cerr << "Unexpected end of .wav file, unable to synchronize!\n";
        exit(1);
    }

	return returnVal;
}

void BallyWavIn::syncBits(bool throwErr)
{
    int goodSineWavs;
    do
    {
	    do
	    {
		    goodSineWavs = 0;
		    while (goodSineWavs == 0)
		    {
                if (getHalfWav(throwErr) * 2 < midPoint)
                    if (getHalfWav(throwErr) * 2 < midPoint)
				    {
                        goodSineWavs++;
				    }
		    }
			try
			{
				getNextByte();
			}
			catch (Error::badWavLength)
			{
				byteStream = 0;
			}
	    } while (byteStream != 0xa5);
    // If the stop bit is not a 1, keep trying
    } while (getBit() == 0);
}

void BallyWavIn::syncWav(void)
{
	static long lastValue = 0;
	average = 0;
	int halfWavLength;
	int goodSineWavs = 0;
	double sineWaveLengthAvg = 0;

	for (filePosition = 0x2c;
	     filePosition < (bytesPerSecond * .25) + 0x2c;
		 filePosition += bytesPerSample * channels)
    {
        average += (lastValue = getSample());
    }

	average = floor(average/(bytesPerSecond * .5));
    zeroBitLength = 0;
	while (goodSineWavs < 100)
	{
		halfWavLength = getHalfWav(false);
		sineWaveLengthAvg = runningAverage(halfWavLength);

		if (abs(halfWavLength - sineWaveLengthAvg) <= 2)
			goodSineWavs++;
		else
			goodSineWavs = 0;
	}
    oneBitLength = floor(sineWaveLengthAvg * 2 + .5);
	zeroBitLength = floor(sineWaveLengthAvg + .5);
    midPoint = (zeroBitLength + oneBitLength) / 2;
    if (verbosity > 1)
    {
        cout << "oneBitLength = " << oneBitLength << "\n";
	    cout << "zeroBitLength = " << zeroBitLength << "\n";
    }
    syncBits(false);
}

unsigned int BallyWavIn::getByteHeader()
{
	byteStream = 0;
	return decodeBits(2,0);
}

unsigned int BallyWavIn::getBit()
{
    byteStream = 0;
    return decodeBits(1,0);
}