// WavDataIn.cpp: implementation of the WavDataInput class.
//
//////////////////////////////////////////////////////////////////////

#ifdef _AFXDLL
    #include "stdafx.h"
#endif
#include "WavDataIn.h"
#include "Error.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

WavDataIn::WavDataIn()
{

}

WavDataIn::~WavDataIn()
{

}

int WavDataIn::open(char *filename, int sign, int pMinHalfWavLength, int pVerbosity)
{
    verbosity = pVerbosity;
    WavFile::openInput(filename);
	wavSign = sign;
	minHalfWavLength = pMinHalfWavLength;
	return 0;
}

bool WavDataIn::passedZero()
{
	static long lastValue = 0;
	long newValue = 0;
	bool retVal;

    newValue = getSample();
	retVal = false;
	if (lastValue < average)
	{
		if (newValue > average)
			retVal = true;
	}
	else if (lastValue > average)
	{
		if (newValue < average)
			retVal = true;
	}
	else if (lastValue == average)
	{
		if (newValue != average)
			retVal = true;
	}

	lastValue = newValue;
	return retVal;
}

int WavDataIn::getHalfWav(bool throwErr)
{
	static int halfWavLength = 0;
	int tempWavLength = 0;
	int lMinHalfWavLength;
	double fpos = filePosition; //ifsWav.tellg();

	// For Ballybin
	if (minHalfWavLength == 0)
		lMinHalfWavLength = zeroBitLength / 3;
	else
		lMinHalfWavLength = minHalfWavLength;

	while (!passedZero() || halfWavLength < lMinHalfWavLength)
		halfWavLength++;

	if (oneBitLength > 0 && halfWavLength > oneBitLength + zeroBitLength)
	{
		if (throwErr)
		{
			ifsWav.seekg(fpos);
			filePosition = fpos;
			throw Error::badWavLength();
		}
	}
	tempWavLength = halfWavLength;
	halfWavLength = 1;
	return tempWavLength;
}

int WavDataIn::getWavLength(void)
{
	return getHalfWav(true) + getHalfWav(true);
}

double WavDataIn::runningAverage(long currentVal)
{
	static long valList[50];
	static int listPtr = 0;
	static int listCount = 0;
	static double listTotal = 0;

	if (listCount > 49)
		listTotal -= valList[listPtr];
	else
		listCount++;

	valList[listPtr++] = currentVal;
	listTotal += currentVal;

	if (listPtr == 50)
		listPtr = 0;

	return listTotal/listCount;
}

unsigned int WavDataIn::decodeBits(int bits, int ReadMode)
{
	static int bit;
	int curWavLength = 0;

	for (int bitCount=0; bitCount<bits ;)
	{
		curWavLength = getWavLength();
        if (curWavLength < midPoint)
        {
			bit = 1;
			byteStream = (byteStream << 1);
			bitCount++;
		}
		else
		{
			bit = 0;
			byteStream = (byteStream << 1) + 1;
			bitCount++;
		}
	}
	return byteStream;
}

unsigned int WavDataIn::getNextBit()
{
	byteStream = 0;
    return decodeBits(1,0);
}

unsigned int WavDataIn::getNextByte()
{
    return decodeBits(8,0);
}
