#ifdef _AFXDLL
    #include "stdafx.h"
#endif
#include "WavOutFile.h"

WavOutFile::WavOutFile(char *outFile) {

	openWavOutFile(outFile);
	zeroBitLength = 7;
	oneBitLength = 14;
	bitcount = 0;
}

WavOutFile::~WavOutFile()
{
	long flength, wavFlength;
	unsigned int lengthLword[4];
	int i;

	wavFlength = (double)ofsWavOutFile.tellp() - (double)8;
	flength = wavFlength - 36;
	ofsWavOutFile.seekp(0x28,ios::beg);
    lengthLword[3] = 0;
    if (flength > 65535)
    {
        lengthLword[2] = (int)floor((double)(flength / 65536));
        flength -= lengthLword[2] * 65536;
    }
    if (flength > 256)
    {
        lengthLword[1] = (int)floor((double)(flength / 256));
        flength -= lengthLword[1] * 256;
    }
    lengthLword[0] = (unsigned int)flength;

    for(i=0;i<4;i++)
		ofsWavOutFile.put((unsigned char)lengthLword[i]);

    /* The file length for the RIFF chunk of the .wav file is in wavFlength */
    /* Write the file length in hi-byte/lo-byte, hi-word/lo-word format to the .wav file */

	ofsWavOutFile.seekp(0x04,ios::beg);
    lengthLword[3] = 0;
    if (wavFlength > 65535)
    {
        lengthLword[2] = (unsigned int)floor((double)(wavFlength / 655306));
        wavFlength -= lengthLword[2] * 65536;
    }
    if (wavFlength > 256)
    {
        lengthLword[1] = (unsigned int)floor((double)(wavFlength / 256));
        wavFlength -= lengthLword[1] * 256;
    }
    lengthLword[0] = (unsigned int)wavFlength;
    for(i=0;i<4;i++)
		ofsWavOutFile.put((unsigned char)lengthLword[i]);
	ofsWavOutFile.close();
	cout << "bitcount was " <<bitcount << "\n";
}

int WavOutFile::openWavOutFile(char *filename)
{
	unsigned char header[2][80]=
	{
		{'R','I','F','F',0x80,0x80,0x80,0x80,'W','A','V','E','f','m','t',' ',0x10,0,0,0,1,0,1,0,0x22,0x56,0,0,0x22,0x56,0,0,1,0,8,0,'d','a','t','a',0,0,0,0,'\0' },
		{'R','I','F','F',0x80,0x80,0x80,0x80,'W','A','V','E','f','m','t',' ',0x10,0,0,0,1,0,1,0,0x44,0xac,0,0,0x44,0xac,0,0,1,0,8,0,'d','a','t','a',0,0,0,0,'\0' }
	};

	ofsWavOutFile.open(filename,ios::out|ios::binary|ios::ate|ios::trunc);
	if (ofsWavOutFile == 0) {
		cerr << "Unable to open " << filename << "\n";
		exit(1);
	}
	ofsWavOutFile.write((char *)header[0],44);
    bytesPerSecond = 22050;

	return 0;
}

void WavOutFile::putBit(int bitLength)
{
    int i;
    int isOdd;
    int halfBitLength;

	bitcount++;
    halfBitLength = (int)(floor((float)bitLength)/2.0);
    isOdd = (int)(ceil((float)bitLength/2.0) - halfBitLength);
    for (i=0;i<halfBitLength;i++)
        ofsWavOutFile.put((unsigned char)0x06);
    if (isOdd)
	{
        ofsWavOutFile.put((unsigned char)0x80);
		i++;
	}
    for (;i<bitLength;i++)
        ofsWavOutFile.put((unsigned char)0xfc);
}

void WavOutFile::putByteHeader(int header)
{
    int i;
    for (i=0; i<2; i++)
    {
        if (header & 2)
	       putBit(oneBitLength);
	    else
	       putBit(zeroBitLength);
        header = header << 1;
    }
}

void WavOutFile::putByte(unsigned char inByte)
{
	int i;

	for (i=0; i<8; i++)
	{
		if (inByte >= 128)
			putBit(oneBitLength);
		else
			putBit(zeroBitLength);
		inByte = inByte << 1;
	}
}

void WavOutFile::putSilence(double silenceLength)
{
    for (int i=0; i<(int)(silenceLength * bytesPerSecond); i++)
        ofsWavOutFile.put((unsigned char)0x00);
}

void WavOutFile::putWord(int header, unsigned char inByteOne, unsigned char inByteTwo)
{
	unsigned int outByteOne;
	unsigned int outByteTwo;
	/* inByteOne  = A7 - A0
	   inByteTwo  = B7 - B0
	   outByteOne = X7 - X0
	   outByteTwo = Y7 - Y0

	   X0 = B7	Y0 = B6 | A0 = Y7  B0 = Y6
       X1 = A7	Y1 = A6 | A1 = X7  B1 = X6
	   X2 = B5	Y2 = B4 | A2 = Y5  B2 = Y4
	   X3 = A5	Y3 = A4 | A3 = X5  B3 = X4
	   X4 = B3	Y4 = B2 | A4 = Y3  B4 = Y2
	   X5 = A3	Y5 = A2 | A5 = X3  B5 = X2
	   X6 = B1	Y6 = B0 | A6 = Y1  B6 = Y0
	   X7 = A1	Y7 = A0 | A7 = X1  B7 = X0
    */
	outByteOne = ((inByteOne & 0x02) > 0)*128 + ((inByteTwo & 0x02) > 0)*64 +
		         ((inByteOne & 0x08) > 0)*32  + ((inByteTwo & 0x08) > 0)*16 +
			     ((inByteOne & 0x20) > 0)*8   + ((inByteTwo & 0x20) > 0)*4  +
			     ((inByteOne & 0x80) > 0)*2   + ((inByteTwo & 0x80) > 0);

	outByteTwo = ((inByteOne & 0x01) > 0)*128 + ((inByteTwo & 0x01) > 0)*64 +
			     ((inByteOne & 0x04) > 0)*32  + ((inByteTwo & 0x04) > 0)*16 +
			     ((inByteOne & 0x10) > 0)*8   + ((inByteTwo & 0x10) > 0)*4  +
			     ((inByteOne & 0x40) > 0)*2   + ((inByteTwo & 0x40) > 0);

    putByteHeader(header);
	putByte(outByteOne);
    putByteHeader(header);
	putByte(outByteTwo);
}
