// Bugs:  15-NOT_archived.bml gets incorrect checksum
#ifdef _AFXDLL
    #include "stdafx.h"
#endif

#include "BallyBinFile.h"
#include "Error.h"
#include "WordVal.h"
#include "BallyWavIn.h"

BallyBinFile::BallyBinFile(void)
{
    createBinFile = false;
    loadCount = 0;
	loadNumber = 0;
    multiLoadNumber = 0;
    setTapeName("");
	pHeaderLength = -1.0;
	pFooterLength = -1.0;
	pGapLength    = -1.0;
    setMinHalfWavLength(0);
    basicPointer = 0;
    setFirstDataLoad(2);
	setScreenBmpName("");

/*
    This is the list of tokens that are used by basic as well the
    characters that the PC can not represent.  The 1st value is the
    character that represents the token, the 2nd value is the
    replacement string used by ballybin, and the 3rd value indicates
    how and when the token is considered as described below:

    0 - This is a BASIC token.  When this type of token appears outside
        of quotes ballybin uses the string exactly as it appears below.
        When this token appears inside quotes, the string is enclosed in
        two pipe characters (the '|' character).  This is to distinguish
        between the case, for example, when the word 'TO' appears in a
        quoted string and the case when the basic token for 'TO ' appears.
        If no distinction were made, astrowav would not know how to encode
        the string properly.
        Notice that most of the BASIC tokens have spaces after them, this
        is part of the TOKEN.  The INPUT token does NOT have a space after it.

    1 - This is a character that either cannot be represented properly
        in the ASCII character set.

    2 - This is a character for which he Bally uses a different value.  The
        ASCII value is invalidated so that there are not two instances of the
        character in the look-up table.
*/

    static struct basicTokenType localBasicTokens[] =
	{
		{31,   "|erase|",      1},
        {91,   "[",            1},
        {92,   "\\",           1},
        {93,   "]",            1},
        {94,   "|up_arrow|",   1},
        {95,   "|left_arrow|", 1},
        {96,   "|down_arrow|", 1},
        {97,   "|right_arrow|",1},
        {98,   "|*|",          1},
        {99,   "|divided_by|", 1},
        {100,  "|graphic_1|",  1},
        {101,  "|graphic_2|",  1},
        {102,  "|graphic_3|",  1},
        {103,  "|graphic_4|",  1},
        {104,  "LIST ",        0},
        {105,  "CLEAR ",       0},
        {106,  "RUN ",         0},
        {107,  "NEXT ",        0},
        {108,  "LINE ",        0},
        {109,  "IF ",          0},
        {110,  "GOTO ",        0},
        {111,  "GOSUB ",       0},
        {112,  "RETURN ",      0},
        {113,  "BOX ",         0},
        {114,  "FOR ",         0},
        {115,  "INPUT",        0}, // No space after token
        {116,  "PRINT ",       0},
        {117,  "STEP ",        0},
        {118,  "RND ",         0},
        {119,  "TO ",          0},
        {0xf8, "?",            0},
		// Mark the end of the array
        {0x00, "zzz",          0}
    };

	// Set all values in the token lists to their normal values
    for (int i = 0; i < 256; i++)
    {
        basicTokens[i][0] = i;
        basicTokens[i][1] = '\0';
    }

    for (i = 0; i < 256; i++)
    {
        characterTokens[i][0] = i;
        characterTokens[i][1] = '\0';
    }

	// Replace the normal character values with Bally specific values
    for (i = 0;; i++)
	{
	// See if we reached the end of the token array
		if (localBasicTokens[i].value == 0x00)
			break;
		// Determine if we need to add pipes and then copy the token to the characterTokens array
        if (localBasicTokens[i].inQuotes)
        {
            strcpy(characterTokens[localBasicTokens[i].value], localBasicTokens[i].mnemonic);
        } else {
            sprintf(characterTokens[localBasicTokens[i].value], "|%s|", localBasicTokens[i].mnemonic);
        }
		// Set up the basic tokens array
        strcpy(basicTokens[localBasicTokens[i].value], localBasicTokens[i].mnemonic);
	}

	// Set up all unprintable characters
	for (i=0; i<=255; i++)
	{
		if (i <= 30 || i >=120)
		{
			sprintf(characterTokens[i], "|%d|", i);
			strcpy(basicTokens[i], characterTokens[i]);
		}
	}
}

BallyBinFile::~BallyBinFile()
{
}

// Close and clean up an output file
void BallyBinFile::close()
{
    if (ofsBallyBinFile)
    {
        if (ofsBallyBinFile->is_open())
        {
            if (tagQue.size())
            {
                if (tagQue.front().name == "DATA")
                    closeDataTag();
                if (tagQue.front().name == "LOAD")
                    closeLoadTag();
                if (tagQue.front().name == "BALLY")
                    closeBallyTag();
            }

            ofsBallyBinFile->close();
        }
    }
    if (ifsBallyBinFile.is_open())
    {
        ifsBallyBinFile.close();
    }
    loadNumber = 0;
    loadCount = 0;
}

int BallyBinFile::getShowScreen()
{
    return showScreen;
}

void BallyBinFile::setShowScreen(int pShowScreen)
{
    showScreen = pShowScreen;
}

int BallyBinFile::getLastBasicLoad()
{
    return lastBasicLoad;
}

void BallyBinFile::setLastBasicLoad(int pLastBasicLoad)
{
    lastBasicLoad = pLastBasicLoad;
}

int BallyBinFile::getFirstDataLoad()
{
    return firstDataLoad;
}

void BallyBinFile::setFirstDataLoad(int pFirstDataLoad)
{
    firstDataLoad = pFirstDataLoad;
}

int BallyBinFile::getMinHalfWavLength()
{
    return minHalfWavLength;
}

void BallyBinFile::setMinHalfWavLength(int pMinHalfWavLength)
{
    minHalfWavLength = pMinHalfWavLength;
}

int BallyBinFile::getVerbosity()
{
    return verbosity;
}

void BallyBinFile::setVerbosity(int pVerbosity)
{
    verbosity = pVerbosity;
}

// Convert a wav file into a data file
int BallyBinFile::decodeWavFile(char *inputName)
{
	double headerLength    = 0,
           dataLength      = 0,
           footerLength    = 0,
           newHeaderLength = 0;
           
	bool noGapExists  = true;

    int  loadByteNum  = 0,
         loadFormat   = 0,
         startBitOne  = 0,
         startBitTwo  = 0;


    unsigned char calcedCheckSum = 0xa5,
                  readCheckSum   = 0,
                  inByteOne      = 0,
                  inByteTwo      = 0,
                  outByteOne     = 0,
                  outByteTwo     = 0;

    BallyWavIn inFile;
    TapeLoad   *tLoad;

    // Open up the Bally audio file
    inFile.open(inputName, 1, getMinHalfWavLength(), verbosity);

    // Reset various timers
    WavTimer headerTimer(&inFile);
    WavTimer dataTimer(&inFile);
    WavTimer footerTimer(&inFile);
    WavTimer fileTimer(&inFile);

    // Set timers to point to the start of RIFF data section
    headerTimer.set(0x2d);
	fileTimer.set(0x2d);

#ifdef USE_IPL98
	// Construct a new image file
	ipl::CImage Image2(160, 90, 1, 1);

	int xPixel = 0,
	    yPixel = 0;
#endif

    headerLength = headerTimer.currentTime();
    try
    {
        loadByteNum = 0;
        loadCount = 1;
 
        while (1)
        {
            try
            {
				if (!noGapExists)
				{
					if (pHeaderLength < 0)
                    {
						headerLength = headerTimer.currentTime();
                    }
					else
                    {
						headerLength = pHeaderLength;
                    }
                    tLoad->setHeaderLength(headerLength);
				}
                noGapExists = false;

                if (verbosity > 1)
                {
					cout << "--------------------\n";
				}
				if (verbosity)
				{
                    cout << "Decoding load " <<loadCount;
                    cout.flush();
                }

                if (verbosity > 1)
                {
    				printf("\nData starts at %3.8f\n",fileTimer.currentTime());
                }

                if (getFileFormat() == 2)
                {
                    if (loadCount < getFirstDataLoad())
                    {
                        loadFormat = 2;  // Interleaved Format
                    }
                    else
                    {
                        loadFormat = 1;  // Continuous Format
                    }
                }
                else
                {
                    loadFormat = 3;      // Binary Format - always interleaved
                }

                dataTimer.reset();
                loadQue.push_back(TapeLoad());
                tLoad = &(loadQue.back());
                tLoad->setHeaderLength(headerLength);
                tLoad->setLoadFormat(loadFormat);
                inByteTwo = 0;
				while (1) 
				{
                    try
                    {
                        // If the stop bit is 1 the data has ended
                        startBitOne = inFile.getBit();
                        inByteOne = inFile.getNextByte();
                        if (inFile.getBit() != 1)
                        {
                            throw Error::badStopBit();
                        }

                        if (startBitOne == 1)
					    {
						    tLoad->setByteCount(loadByteNum * 2);
                            throw Error::loadComplete();
                        }

                        // Add the previous byte to the checksum
						calcedCheckSum += WordVal::revByte(inByteTwo);
                        // Set the current byte as the POSSIBLE read checksum
						readCheckSum = WordVal::revByte(inByteOne);

                        startBitTwo = inFile.getBit();
					    inByteTwo = inFile.getNextByte();

                        if (!startBitTwo)
                        {
                            // Add the previous byte to the checksum
							calcedCheckSum += WordVal::revByte(inByteOne);
                            // Set the current byte as the POSSIBLE 
							readCheckSum = WordVal::revByte(inByteTwo);
                        }

                        if (inFile.getBit() != 1)
                        {
                            throw Error::badStopBit();
                        }

                        if (loadFormat == 2 || loadFormat == 3)
                        {
                            convertWord(inByteOne, inByteTwo, &outByteOne, &outByteTwo);
					        tLoad->dataBuffer[loadByteNum] = outByteOne;
					        tLoad->dataBuffer[0x1000 + loadByteNum] = outByteTwo;
                        }
                        else if (loadFormat == 1)
                        {
                            tLoad->dataBuffer[loadByteNum * 2] = WordVal::revByte(inByteOne);
                            tLoad->dataBuffer[loadByteNum * 2 + 1] = WordVal::revByte(inByteTwo);
                        }
#ifdef MSDOS
                        if ((loadCount < getFirstDataLoad()) && (loadByteNum < (20 * 90)) && getShowScreen())
                        {
                            char mask[8] = {(char)0x80, (char)0x20, (char)0x08, (char)0x02,
                                            (char)0x40, (char)0x10, (char)0x04, (char)0x01};
                            int i;
                            for (i=0; i<8; i++)
                            {
                                if (outByteTwo & mask[i])
                                {
                                    fprintf(stdout, "%c", (char)PIXEL);
#ifdef USE_IPL98
									if (strlen(screenBmpName))
									{
										Image2.SetPixel(xPixel, yPixel, 0);
									}
#endif
                                }
                                else
                                {
                                    fprintf(stdout, " ");
                                }
#ifdef USE_IPL98
								xPixel++;
#endif
                            }
#ifdef USE_IPL98
							if (xPixel == 160)
							{
								xPixel = 0;
								yPixel++;
							}
#endif
                        }
#endif
						loadByteNum++;

                        if (startBitTwo == 1)
					    {
						    tLoad->setByteCount(loadByteNum * 2 - 1);
                            throw Error::loadComplete();
                        }
                    }

                    catch (Error::endOfFile)
                    {
                        cerr << "  Warning:  Load did not have footer tone\n";
        				footerTimer.reset();
                        throw Error::loadComplete();
                    }

					catch (Error::badWavLength)
					{
						cerr << "  Warning:  File had silence at the end\n";
						footerTimer.reset();
						throw Error::loadComplete();
					}

                    catch (Error::loadComplete)
                    {
        				footerTimer.reset();
                        throw Error::loadComplete();
					}
				}
				footerTimer.reset();
            }

            catch (Error::loadComplete)
            {
                if (getFileFormat() != 1)
                {
                    tLoad->setByteCount(tLoad->getByteCount()-1);
                    loadByteNum = ceil(tLoad->getByteCount() / 2);
                }
                if (calcedCheckSum + readCheckSum == 0xff)
                {
                    if (verbosity == 1)
                    {
                        printf(" OK!\n");
                    }
                } else {
                    fprintf(stderr, "\r    Invalid Checksum: Calculated=%x - Read=%x\n", calcedCheckSum, 0xff - readCheckSum);
                }

                tLoad->setCheckSum(0xff - calcedCheckSum);

                calcedCheckSum = 0xa5;
                footerTimer.reset();
                dataLength = dataTimer.currentTime();
                tLoad->setLoadNumber(loadCount);

			    try // Look for footer length
			    {
					while (inFile.getBit());
					inByteOne = inFile.getNextByte();
					if (inByteOne == 0xa5) {
                        if (inFile.getBit() != 1)
                        {
                            throw Error::badStopBit();
                        };
						noGapExists = true;
						throw Error::endOfFile();
					}

                    if (startBitOne == 0)
					{
						if (inFile.getNextByte() == 0xa5)
                        {
                            if (inFile.getBit() != 1)
                            {
                                throw Error::badStopBit();
                            };
							noGapExists = true;
                        }
					}
                }

                catch (Error::endOfFile)
                {
                }

				catch (Error::badWavLength)
				{
				}

				if (pFooterLength < 0)
					footerLength = footerTimer.currentTime();
				else
					footerLength = pFooterLength;

				if (noGapExists)
				{
					if (pFooterLength < 0)
						footerLength = footerLength / 2;
					if (pHeaderLength < 0)
						newHeaderLength = footerLength;
					if (pGapLength < 0)
						tLoad->setGapLength(0);
				}

                tLoad->setFooterLength(footerLength);

                footerTimer.reset();
			    headerTimer.reset();

				if (!noGapExists)
				{
					try // Look for gap
					{
						int i = 0;
						try {
							while (i < 10)
							{
								inByteOne = inFile.decodeBits(1,0);
								if (inByteOne == 0xff)
								{
									i++;
								}
								else
								{
									headerTimer.reset();
									i = 0;
								}
							}
						}
						catch (Error::badWavLength)
						{
							int lHalfWavLength = 0;
							int lMinHalfWavLength;

							// For Ballybin
							if (minHalfWavLength == 0)
								lMinHalfWavLength = inFile.zeroBitLength / 3;
							else
								lMinHalfWavLength = minHalfWavLength;

							while (!inFile.passedZero() || lHalfWavLength < lMinHalfWavLength)
								lHalfWavLength++;
							headerTimer.reset();
						}
						catch (Error::endOfFile)
						{
						}
						if (pGapLength < 0)
							tLoad->setGapLength(footerTimer.currentTime() - headerTimer.currentTime());
						else
							tLoad->setGapLength(pGapLength);
					}

					catch (Error::endOfFile)
					{
					}
				}

                if (verbosity > 1)
                {
    				printf("Data ends at %3.8f\n", fileTimer.currentTime());
    				printf("Data length is %3.8f\n",dataLength);
					tLoad->toString();
                }

                if (loadFormat == 2 || loadFormat == 3)
                {
					// Compress
				    for (int i=0; i<loadByteNum; i++)
					    tLoad->dataBuffer[i+loadByteNum] = tLoad->dataBuffer[i+0x1000];
					tLoad->dump();
                }
                if (lastBasicLoad >= loadCount)
                {
                    setBasicFlag();
                } else {
                    clearBasicFlag();
                }

                headerLength = newHeaderLength;
                loadByteNum = 0;
            }
			if (inFile.eof())
            {
				throw Error::conversionComplete();
            }
			else
			{
				if (!noGapExists)
				{
					try // Look for next load
					{
						inFile.syncBits(false);
					}
					catch (Error::endOfFile)
					{
						throw Error::conversionComplete();
					}
				}
			}
			loadCount++;
        }
	    cout.setf(ios::dec,ios::basefield);
        throw Error::conversionComplete();
    }

    catch(Error::endOfFile)
    {
        cerr << "Unexpected end of input reached\n";
        return 1;
    }

	catch(Error::badOutputFileName)
	{
		cerr << "Invalid output .wav file name\n";
		return 1;
	}

    catch(Error::badStopBit)
    {
        cerr << "Bad stop bit\n";
        return 1;
    }

    catch(Error::conversionComplete)
    {
#ifdef USE_IPL98

		if (Image2.Save(screenBmpName) != 1)
		{
			cerr << "Unable to save screen bitmap " << screenBmpName << endl;
		}
#endif
        return 0;
    }
}

void BallyBinFile::writeBinFile(char *binName)
{
    // Open the BallyBinFile in write mode
    open(binName,"w",getVerbosity());

    // Start putting out the binary file
	putBallyTag(loadQue.size());
    if (strlen(tapeName))
    {
	    putNameTag(tapeName);
    }
    int loadCount = 1;
    while (loadQue.size())
    {
        if (getFileFormat() == 2)
        {
            // The number of BASIC loads is stored in lastBasicLoad
            if (getLastBasicLoad() >= loadCount)
            {
                setBasicFlag();
            }
            else
            {
                clearBasicFlag();
            }

            int lLoadFormat;
            if (loadCount < getFirstDataLoad())
            {
                lLoadFormat = 2; // Screen/Data interleaved format
            }
            else
            {
                lLoadFormat = 1; // Continuous Data format
            }
            // Convert the load
            loadQue.front().convertLoadFormat(lLoadFormat);
            // Write the converted load
            putLoad(lLoadFormat);
        }
        else
        {
            loadQue.front().convertLoadFormat(3);
            putLoad(3);      // Binary Format - always interleaved
        }
        // Get the next load to the front of the queue
        loadQue.pop_front();
        loadCount++;
    }
    close();
}

void BallyBinFile::getAttributeVal(char *tag, char *pAttribute, char *attributeVal)
{
    char lTag[256];
    char *attribute, *lAttributeVal;
    strcpy(attributeVal, "");
    strcpy(lTag, tag);

    strtok(lTag, " \t>");
    attribute = strtok(NULL, " =");
    while (attribute > 0) {
        lAttributeVal = strtok(NULL, " \t>");
        if (!stricmp(attribute, pAttribute)) {
            strcpy(attributeVal, lAttributeVal);
            break;
        } else {
            attribute = strtok(NULL, " =");
        }
    }
    if (strlen(attributeVal) == 0)
        strcpy(attributeVal, "0");
}

int BallyBinFile::processOpenTagData(char *token, char *tag, char *fullTag)
{
    char attribVal[80];
    unsigned char byteBuffer[1];
    unsigned char wordBuffer[1];
	if (!strcmp(token, "BALLY"))
	{
        getAttributeVal(fullTag, "format", attribVal);
        setFileFormat(atoi(attribVal));
        if (fileFormat == 1)
        {
			ifsBallyBinFile.read(byteBuffer, 1);
			loadCount = (int)byteBuffer[1];
			ifsBallyBinFile.read(wordBuffer, 2);
			fileLength = WordVal::wordVal(wordBuffer);
			ifsBallyBinFile.read(wordBuffer, 2);
        } else if (fileFormat == 2) {
            getAttributeVal(fullTag, "loadCount", attribVal);
            loadCount = atoi(attribVal);
            getAttributeVal(fullTag, "fileLength", attribVal);
            fileLength = atof(attribVal);
        } else {
            cerr << "Invalid file format " << fileFormat << " found\n";
            exit (1);
        }
		while(processTag("BALLY"));
		return 0;
	}
    return -1;
}

int BallyBinFile::processBallyTagData(char *token, char *tag, char *fullTag)
{
    char attribVal[80];
    unsigned char byteBuffer[1];
	if (!strcmp(token, "/BALLY"))
    {
		return 0;
    } else if (!strcmp(token, "NAME")) {
		getTagValue(token, tapeName);
        setTapeName(tapeName);
		return 1;
    } else if (!strcmp(token, "LOAD")) {
        // Add a new TapeLoad to the end of the que and get a pointer to it
        loadQue.push_back(TapeLoad());
        currLoad = &(loadQue.back());

        // Get a pointer to the data buffer of the TapeLoad
        loadBuffer = currLoad->dataBuffer;

        // Determine the format of the TapeLoad (not currently used)
        getAttributeVal(fullTag, "format", attribVal);
        int loadFormat = atoi(attribVal);
        currLoad->setLoadFormat(loadFormat);
        if (loadFormat == 3)
        {
			ifsBallyBinFile.read(byteBuffer, 1);
            currLoad->setLoadNumber((int)*byteBuffer);
        } else if (loadFormat == 2 || loadFormat == 1) {
            int hexVal;
            getAttributeVal(fullTag, "loadNumber", attribVal);
            currLoad->setLoadNumber(atoi(attribVal));
            getAttributeVal(fullTag, "headerLength", attribVal);
			if (atof(attribVal) == 0)
			{
				strcpy(attribVal, "1.0");
			}
            currLoad->setHeaderLength(atof(attribVal));
            getAttributeVal(fullTag, "footerLength", attribVal);
			if (atof(attribVal) == 0)
			{
				strcpy(attribVal, "0.25");
			}
            currLoad->setFooterLength(atof(attribVal));
            getAttributeVal(fullTag, "gapLength", attribVal);
            currLoad->setGapLength(atof(attribVal));
            getAttributeVal(fullTag, "checkSum", attribVal);
            if (!strcmp(attribVal, "0"))
                hexVal = 0;
            else
                sscanf(attribVal, "0x%2x", &hexVal);
            currLoad->setCheckSum(hexVal);
        } else {
            cerr << "Invalid load format " << loadFormat << " found in load " << currLoad->getLoadNumber();
            exit (1);
        }
		while (processTag("LOAD"));
		return 1;
	}
    return -1;
}

int BallyBinFile::processLoadTagData(char *token, char *tag, char *fullTag)
{
    char attribVal[80];
    unsigned char wordBuffer[1];

	if (!strcmp(token, "HEADER"))
	{
        if (currLoad->getLoadFormat() == 3)
        {
    		ifsBallyBinFile.read(wordBuffer, 2);
			if (pHeaderLength < 0)
				currLoad->headerLength = ((double)wordBuffer[0] + (double)wordBuffer[1] / 100);
			else
				currLoad->headerLength = pHeaderLength;
        } else if (currLoad->getLoadFormat() == 2 || currLoad->getLoadFormat() == 1) {
			if (pHeaderLength < 0)
            {
                getAttributeVal(fullTag, "headerLength", attribVal);
                currLoad->headerLength = atof(attribVal);
            } else {
				currLoad->headerLength = pHeaderLength;
            }
        }
        if (verbosity > 1)
        {
			cout << "    Header length: " << currLoad->headerLength << " seconds \n";
        }

		processTag("HEADER");
		return 1;
    } else if (!strcmp(token, "FOOTER")) {
        if (currLoad->getLoadFormat() == 3)
        {
    		ifsBallyBinFile.read(wordBuffer, 2);
			if (pFooterLength < 0)
				currLoad->footerLength = ((double)wordBuffer[0] + (double)wordBuffer[1] / 100);
			else
				currLoad->footerLength = pFooterLength;
        } else if (currLoad->getLoadFormat() == 2 || currLoad->getLoadFormat() == 1) {
			if (pFooterLength < 0)
            {
                getAttributeVal(fullTag, "footerLength", attribVal);
                currLoad->footerLength = atof(attribVal);
            }
			else
				currLoad->footerLength = pFooterLength;
        }
        if (verbosity > 1)
        {
			cout << "    Footer Length: "<<currLoad->footerLength<<" seconds\n";
        }

		processTag("FOOTER");
		return 1;
    } else if (!strcmp(token, "GAP")) {
        if (currLoad->getLoadFormat() == 3)
        {
    		ifsBallyBinFile.read(wordBuffer, 2);
			if (pGapLength < 0)
				currLoad->setGapLength((double)wordBuffer[0] + (double)wordBuffer[1] / 100);
			else
				currLoad->setGapLength(pGapLength);
        } else if (currLoad->getLoadFormat() == 2 || currLoad->getLoadFormat() == 1) {
            getAttributeVal(fullTag, "gapLength", attribVal);
            currLoad->setGapLength(atof(attribVal));
        }

        if (verbosity > 1)
        {
			cout << "    Gap Length: "<<currLoad->getGapLength()<<" seconds\n";
        }

		processTag("GAP");
		return 1;
	} else if (!strcmp(token, "DATA")) {
        if (currLoad->getLoadFormat() == 3)
        {
		    ifsBallyBinFile.read(wordBuffer, 2);
			loadLength = WordVal::wordVal(wordBuffer);
			ifsBallyBinFile.read(loadBuffer, ((long)loadLength%2 == 0)?loadLength:loadLength+1);
        } else if (currLoad->getLoadFormat() == 2 || currLoad->getLoadFormat() == 1) {
            getAttributeVal(fullTag, "length", attribVal);
            loadLength = atol(attribVal);
        }
        currLoad->setByteCount(loadLength);

        if (verbosity > 1)
        {
			cout << "    Load length: " << loadLength << " bytes\n";
        }

        if (currLoad->getLoadFormat() == 3)
        {
			processTag("DATA");
            return 1;
        } else if (currLoad->getLoadFormat() == 2 || currLoad->getLoadFormat() == 1) {
            basicPointer = 0;
            while (processTag("DATA"));
            ofstream textFile;
            char screenName[80];
            sprintf(screenName,"screen_%d.txt", currLoad->getLoadNumber());
            textFile.open(screenName, ios::out|ios::trunc);
            char mask[8] = {(char)0x80, (char)0x20, (char)0x08, (char)0x02,
                            (char)0x40, (char)0x10, (char)0x04, (char)0x01};

            int i, j, k;
            for (i=ceil(loadLength / 2); i<ceil(loadLength / 2) + (20 * 102);)
            {
                for (j=0; j<20; j++)
                {
                    for (k=0; k<8; k++)
                    {
                        if (loadBuffer[i] & mask[k])
                        {
                            textFile << "*";
                        }
                        else
                        {
                            textFile << " ";
                        }
                    }
                    i++;
                }
                textFile << "\n";
            }

            textFile.close();
            return 1;
        }
    } else if (!strcmp(token, "/LOAD")) {
		loadNumber++;
        long lastByte = currLoad->getMaxBytePointer();
        if ((currLoad->getByteCount() < (basicPointer) * 2) && currLoad->getLoadFormat() == 2)
        {
            currLoad->setByteCount(((basicPointer) * 2));
        }
		return 0;
    }
    return -1;
}

int BallyBinFile::processHexTag(char *token, char *tag, char*fullTag)
{
    long address;
    char attribVal[80];

    getAttributeVal(fullTag, "count", attribVal);
    hexLength = atoi(attribVal);
    getAttributeVal(fullTag, "address", attribVal);
    sscanf(attribVal, "%4x", &address);
    char charBuffer[512], currentChar;
    int i = 0;
    // Get the hex string
    for (i=0; i<512; i++)
    {
        ifsBallyBinFile.read(&currentChar, 1);
        if (currentChar == '<')
        {
            charBuffer[i]=0;
            ifsBallyBinFile.seekg(-1, ios::cur);
            break;
        }
        charBuffer[i] = currentChar;
    }
	charBuffer[i] = '\0';
    char *hexByte;
    int m;
    char n;
    hexByte = strtok(charBuffer, " ");
    hexLength = 0;
    for (m=0;; m++)
    {
        sscanf(hexByte, "%x", &n);
        loadBuffer[m+address] = n;
        currLoad->setMaxBytePointer(m+address);
        hexByte = strtok(NULL, " ");
        hexLength++;
        if (hexByte == NULL)
            break;
    }
	while(processTag("HEX"));
    return 1;
}

int BallyBinFile::processBrTag()
{
    char currentChar;
    int i, j;
    char closeBr[] = "/BR>";
	long currentPosition = 0;

    for (i=0; i<8000; i++)
    {
        ifsBallyBinFile.read(&currentChar, 1);
		// Find the end token
		if (currentChar == '<')
		{
            currentPosition = ifsBallyBinFile.tellg();
            for (j=0; j<4; j++)
            {
                ifsBallyBinFile.read(&currentChar, 1);
                if (currentChar != closeBr[j])
                {
                    currentChar = '<';
                    break;
                }
            }
            if (j == 4)
            {
                break;
            } else
            {
                ifsBallyBinFile.seekg(currentPosition);
            }
		}
		// if i = 8000 then we had a serious problem
	}
	return 1;
}

int BallyBinFile::processLineTag(char *token, char *tag, char *fullTag)
{
    char attribVal[80], basicLine[8000], currentChar;
    int i, j, inQuotes = 0;;
    long lineNumber = 0, currentPosition = 0;
    char closeLine[] = "/LINE>";
	char openBr[] = "BR>";

    getAttributeVal(fullTag, "number", attribVal);
    lineNumber = atol(attribVal);

    // Get the basic line
    for (i=0; i<8000; i++)
    {
        ifsBallyBinFile.read(&currentChar, 1);
        if (currentChar == '/"')
        {
            if (inQuotes)
                inQuotes = 0;
            else
                inQuotes = 1;
        }

        if (currentChar == '<' && !inQuotes)
        {
            currentPosition = ifsBallyBinFile.tellg();
            for (j=0; j<6; j++)
            {
                ifsBallyBinFile.read(&currentChar, 1);
                if (currentChar != closeLine[j])
                {
                    currentChar = '<';
                    break;
                }
            }
            if (j == 6)
            {
                basicLine[i]=0;
                ifsBallyBinFile.seekg(currentPosition-1);
                break;
            } else
            {
                ifsBallyBinFile.seekg(currentPosition);
            }
            for (j=0; j<3; j++)
            {
                ifsBallyBinFile.read(&currentChar, 1);
                if (currentChar != openBr[j])
                {
                    currentChar = '<';
                    break;
                }
            }
            if (j == 3)
            {
				processBrTag();
				ifsBallyBinFile.read(&currentChar, 1);
            } else
            {
                ifsBallyBinFile.seekg(currentPosition);
            }
        }
        basicLine[i] = currentChar;
    }
    encodeBasicLine(loadBuffer, &basicPointer, lineNumber, basicLine);
    while(processTag("LINE"));
    return 1;
}

int BallyBinFile::processBasicTag(char *token, char *tag, char *fullTag)
{
    int returnVal = 0, endByte = 0;
	char attribVal[80];

	getAttributeVal(fullTag, "endByte", attribVal);
	basicEndByte = atoi(attribVal);
    while (processTag("BASIC"));
    return 1;
}

int BallyBinFile::processDataTagData(char *token, char *tag, char *fullTag)
{
    int returnVal = 0;

    if (!strcmp(token, "/DATA"))
        returnVal = 0;
    else if (!strcmp(token, "HEX"))
        returnVal = processHexTag(token, tag, fullTag);
    else if (!strcmp(token, "BASIC"))
    {
        returnVal = processBasicTag(token, tag, fullTag);
    }
    return returnVal;
}

int BallyBinFile::processTag(char *parentTag)
{
	char *token;
    char fullTag[256];
    char tag[512];
    int index;
    unsigned char charBuffer = 0;

    while (charBuffer != '<')
    {
	    ifsBallyBinFile.read(&charBuffer, 1);
    }
	index = 0;
	ifsBallyBinFile.read(&charBuffer, 1);
	while (charBuffer != '>')
	{
		tag[index++] = charBuffer;
		ifsBallyBinFile.read(&charBuffer, 1);
	}
	tag[index++] = '\0';

    strcpy(fullTag, tag);
	token = strtok(tag, " \t>>");
    int returnVal = -1;

	if (!strcmp(parentTag, "OPEN"))
	{
        returnVal = processOpenTagData(token, tag, fullTag);
	}
	else if (!strcmp(parentTag, "BALLY"))
	{
        returnVal = processBallyTagData(token, tag, fullTag);
	}
	else if (!strcmp(parentTag, "FOOTER"))
	{
		if (!strcmp(token, "/FOOTER"))
            returnVal = 0;
	}
	else if (!strcmp(parentTag, "GAP"))
	{
		if (!strcmp(token, "/GAP"))
            returnVal = 0;
	}
	else if (!strcmp(parentTag, "LOAD"))
	{
        returnVal = processLoadTagData(token, tag, fullTag);
	}
	else if (!strcmp(parentTag, "HEADER"))
	{
		if (!strcmp(token, "/HEADER"))
            returnVal = 0;
	}
	else if (!strcmp(parentTag, "DATA"))
	{
        returnVal = processDataTagData(token, tag, fullTag);
    }
	else if (!strcmp(parentTag, "BASIC"))
	{
        if (!strcmp(token, "/BASIC"))
        {
// Change to use new BASIC attribute to set 00 value
            loadBuffer[basicPointer++] = basicEndByte; //0x00;
            loadBuffer[basicPointer++] = 0xff;
            currLoad->setMaxBytePointer((basicPointer) * 2);
            returnVal = 0;
        }
        else if (!strcmp(token, "LINE"))
            returnVal = processLineTag(token, tag, fullTag);
    }
    else if (!strcmp(parentTag, "HEX"))
    {
        if (!strcmp(token, "/HEX"))
            returnVal = 0;
        else if (!strcmp(token, "CHAR"))
        {
            char charBuffer[256];
            ifsBallyBinFile.read(charBuffer, hexLength);
            processTag("CHAR");
            returnVal = 1;
        }
    }
    else if (!strcmp(parentTag, "LINE"))
    {
        if (!strcmp(token, "/LINE"))
            returnVal = 0;
    }
    else if (!strcmp(parentTag, "CHAR"))
    {
        if (!strcmp(token, "/CHAR"))
            returnVal = 0;
    }

    if (returnVal >= 0)
        return returnVal;

	cerr << "Invalid Tag " << token << " found for parent tag " << parentTag << "\n";
	exit(1);

	return 0;
}

void BallyBinFile::getTagValue(char *endTag, char *tagValue)
{
	char charBuffer[256];
	int index = 0;

	ifsBallyBinFile.read(charBuffer, 1);
	while (charBuffer[0] != '<')
	{
		tagValue[index++] = charBuffer[0];
		ifsBallyBinFile.read(charBuffer, 1);
	}

	ifsBallyBinFile.read(charBuffer, 1);

	if (charBuffer[0] != '/')
	{
		cerr << "Tag " << endTag << " not properly ended: No slash found in end-tag\n";
		exit(1);
	}

	ifsBallyBinFile.read(charBuffer, strlen(endTag));
	charBuffer[strlen(endTag)] = '\0';

	if (strcmp(charBuffer, endTag))
	{
		cerr << "Tag " << endTag << " not properly ended: Found " << charBuffer << " instead\n";
		exit(1);
	}

	ifsBallyBinFile.read(charBuffer, 1);

	if (charBuffer[0] != '>')
	{
		cerr << "Tag " << endTag << " not properly ended: closing bracket found on end-tag\n";
		exit(1);
	}

	tagValue[index++] = '\0';
}

int BallyBinFile::open(char *fileName, char *mode, int pVerbosity)
{
    verbosity = pVerbosity;
    if (!strcmp(mode, "r"))
    {
	    ifsBallyBinFile.open(fileName, ios::in|ios::binary|ios::ate|ios::nocreate);
	    if (ifsBallyBinFile == 0)
        {
		    cerr << "Unable to open " << fileName << "\n";
		    exit(1);
	    }
      	ifsBallyBinFile.seekg(0);
        processTag("OPEN");
    }
    else if (!strcmp(mode, "w"))
    {
        cleanFileName(fileName,".bin");
        cleanFileName(fileName,".bml");
        strcpy(binName, fileName);
        tagQue.push_front(Tag("OPEN"));
        createBinFile = true;
    }
    else
    {
        cerr << "Invalid open mode for BallyBinFile " << fileName << "\n";
        exit(1);
    }


	return 0;
}

void BallyBinFile::writeWavFile(char *pWavName, long pSampleRate, int pBitsPerSample, int pChannels)
{
    int lvLoadNumber;
    TapeLoad currTapeLoad;
    unsigned char *loadBuffer;
    char wavName[256];
#ifdef USE_IPL98
	char bmpName[256];
#endif
    strcpy(wavName, pWavName);
    cleanFileName(wavName, ".wav");
	sampleRate = pSampleRate;
	bitsPerSample = pBitsPerSample;
	channels = pChannels;

    for(lvLoadNumber = 1; lvLoadNumber <= loadQue.size(); lvLoadNumber++)
    {
        currTapeLoad = loadQue.front();
		loadLength = (currTapeLoad.getByteCount());
        loadBuffer = currTapeLoad.dataBuffer;
        loadQue.pop_front();
        loadQue.push_back(currTapeLoad);
        long loadOffset = ceil(loadLength / 2);

#ifdef USE_IPL98
		getScreenBmpName(bmpName);
        if (currTapeLoad.getLoadFormat() == 2 && lvLoadNumber == 1 && strlen(bmpName))
		{
			ipl::CImage inputImage;

			if (inputImage.Load(bmpName) != 1)
			{
				cerr << "Unable to load bitmap " << bmpName << endl;
				exit(1);
			};

			inputImage.PrintInfo();
			unsigned char screenByte = 0;

			for (int yPixel = 0; yPixel < 90; yPixel++)
			{
				for (int xByte = 0; xByte < 20; xByte++)
				{
					screenByte = 0;
					for (int xBit = 0; xBit < 8; xBit++)
					{
						char mask[8] = {(char)0x80, (char)0x20, (char)0x08, (char)0x02,
										(char)0x40, (char)0x10, (char)0x04, (char)0x01};
						int xPixel = xByte * 8 + xBit;
						if (inputImage.GetPixel(xPixel, yPixel) == 0)
						{
							screenByte |= mask[xBit];
						} else {
							printf("");
						}
					}
					currTapeLoad.dataBuffer[xByte + (yPixel * 20) + loadOffset] = screenByte;
				}
			}
		}
#endif
        if (lvLoadNumber == 1)
		{
    		char loadName[256];

			sprintf(loadName, "%s.wav", wavName);

			if (currTapeLoad.getLoadNumber() > 1)
            {
				wavFile.close();
            }

			wavFile.open(loadName, sampleRate, bitsPerSample, channels);

            if (verbosity)
			{
                cout << "Creating audio file " << loadName << "\n";
				cout.flush();
			}
        }

        if (verbosity)
		{
			cout << "  Processing load " << lvLoadNumber;
			cout.flush();
		}

        WavTimer headerTimer(&wavFile);

        while (headerTimer.currentTime() < currTapeLoad.getHeaderLength())
        {
            wavFile.putBallyBit(1);
        }

        // Put byte to indicate start of data
        wavFile.checkSum = 0;
        wavFile.cs = 0xa5;
        wavFile.putBallyBit(0);
        wavFile.putByte(0xa5);
        wavFile.putBallyBit(1);

        int startBitTwo;

        for (int i=0; i < loadOffset; i++)
        {
			// Should only happen on odd length loads
            if (i+loadOffset >= loadLength)
            {
                startBitTwo = 1;
            }
            else
            {
                startBitTwo = 0;
            }
            if (i == loadOffset - 1)
            {
	            unsigned char inByteOne, inByteTwo, outByteOne, outByteTwo, readCheckSum, calcCheckSum;
                if (currTapeLoad.getLoadFormat() == 2)
                {
                    BallyBinFile::convertWord(loadBuffer[i], loadBuffer[i+loadOffset], &outByteOne, &outByteTwo);
                }
                else if (currTapeLoad.getLoadFormat() == 3)
                {
                    BallyBinFile::convertWord(loadBuffer[i], loadBuffer[i+loadOffset], &outByteOne, &outByteTwo);
                } else {
                    outByteOne = WordVal::revByte(loadBuffer[i * 2]);
                    outByteTwo = WordVal::revByte(loadBuffer[i * 2 + 1]);
                }

                 calcCheckSum = 0xff - wavFile.cs;
                if (currTapeLoad.getLoadFormat() == 3)
                {
                    if (startBitTwo) {
                        readCheckSum = WordVal::revByte(outByteOne);
                    } else {
                        calcCheckSum -= WordVal::revByte(outByteOne);
                        readCheckSum = WordVal::revByte(outByteTwo);
                    }
                } else if (currTapeLoad.getLoadFormat() == 2 || currTapeLoad.getLoadFormat() == 1) {
					if (startBitTwo) {
						readCheckSum = currTapeLoad.getCheckSum();
					} else {
						calcCheckSum -= outByteOne;
					}
                }

                if ((calcCheckSum == readCheckSum) && (verbosity == 1))
                {
                    printf(" OK!\n");
                }

                if (calcCheckSum != readCheckSum || currTapeLoad.getLoadFormat() == 2 || currTapeLoad.getLoadFormat() == 1)
                {
					if (calcCheckSum != readCheckSum)
                    {
                        fprintf(stderr, "Calculated Checksum = %x - Read Checksum = %x\n", calcCheckSum, readCheckSum);
						cerr << "Repairing checksum\n";
                    }
                    if (startBitTwo)
                    {
                        inByteOne = WordVal::revByte(calcCheckSum);
                        inByteTwo = 0xff;
                    } else {
                        inByteOne = outByteOne;
                        inByteTwo = WordVal::revByte(calcCheckSum);
                    }
                    if (currTapeLoad.getLoadFormat() == 2 || currTapeLoad.getLoadFormat() == 3)
                    {
                        BallyBinFile::convertWord(inByteOne, inByteTwo, &outByteOne, &outByteTwo);
                        wavFile.putBasicWord(startBitTwo, outByteOne, outByteTwo);
                    } else {
                        // not sure if this is correct
                        outByteOne = WordVal::revByte(inByteOne);
                        outByteTwo = WordVal::revByte(inByteTwo);
                        wavFile.putDataWord(startBitTwo, outByteOne, outByteTwo);
                    }
                    break;
                }
            }
            if (currTapeLoad.getLoadFormat() == 2)
            {
                wavFile.putBasicWord(startBitTwo, loadBuffer[i], loadBuffer[i+loadOffset]);
            }
            else if (currTapeLoad.getLoadFormat() == 3) 
            {
                wavFile.putBasicWord(startBitTwo, loadBuffer[i], loadBuffer[i+loadOffset]);
            }
            else if (currTapeLoad.getLoadFormat() == 1)
                wavFile.putDataWord(startBitTwo, loadBuffer[i * 2], loadBuffer[i * 2 + 1]);
        }

        headerTimer.reset();
        while (headerTimer.currentTime() < currTapeLoad.getFooterLength())
        {
	        wavFile.putBallyBit(1);
        }

        headerTimer.reset();
        int bitCount = 2;
		wavFile.putSilence(currTapeLoad.getGapLength());
/*
        while (headerTimer.currentTime() < currTapeLoad.getGapLength())
        {
            if (bitCount < 2)
                wavFile.putBallyBit(0);
            else
                wavFile.putSilence(0);
        }
*/
    }
}

void BallyBinFile::checkParentTag(std::string tagName, std::string parentTagName)
{
    if (tagQue.front().name != parentTagName)
    {
        cerr << "Can not open <"<<tagName.c_str()<<"> tag under <"<<tagQue.front().name.c_str()<<">\n";
        exit(1);
    }
}

void BallyBinFile::verifyTagIsOpen(std::string tagName)
{
    if (tagQue.front().name != tagName)
    {
        cerr << "Can not close <"<<tagName.c_str()<<">, <"<<tagQue.front().name.c_str()<<"> is still open\n";
        exit(1);
    }
}

void BallyBinFile::putBallyTag(int pLoadCount)
{
    char tagBuffer[80];

    checkParentTag("BALLY","OPEN");

	loadCount = 0;
    if (createBinFile)
    {
        if (getFileFormat() == 1)
        {
		    sprintf(loadName, "%s.bin", binName);
        }
        else
        {
            sprintf(loadName, "%s.bml", binName);
        }

        ofsBallyBinFile = new ofstream;

        ofsBallyBinFile->open(loadName ,ios::out|ios::binary|ios::trunc);
        if (fileFormat == 1)
        {
            sprintf(tagBuffer,"<BALLY format=1>");
       	    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
            tagQue.push_front(Tag("BALLY", ofsBallyBinFile->tellp()));
            ofsBallyBinFile->write("\0\0\0\0\0", 5);
        }
        else
        {
            sprintf(tagBuffer,"<BALLY format=%d loadCount=%d", fileFormat, pLoadCount);
       	    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
            tagQue.push_front(Tag("BALLY", ofsBallyBinFile->tellp()));
            sprintf(tagBuffer,"%40.40s"," ");
       	    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
        }
    } else
        tagQue.push_front(Tag("BALLY"));
}

void BallyBinFile::closeBallyTag(void)
{
    char tagBuffer[80];
	unsigned char parmBuffer[3];
	int i = 0;

	double fileLength;

    verifyTagIsOpen("BALLY");
    if (createBinFile)
    {
        sprintf(tagBuffer,"</BALLY>");
        ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
        fileLength = ofsBallyBinFile->tellp();
        ofsBallyBinFile->seekp(tagQue.front().filePosition);

        if (fileFormat == 1)
        {
            putByte(parmBuffer, &i, loadCount);
	        putWord(parmBuffer, &i, fileLength);
	        ofsBallyBinFile->write(parmBuffer, 3);
	        ofsBallyBinFile->seekp(0,ios::end);
        } else {
            sprintf(tagBuffer, " fileLength=%0.0f>", fileLength);
            ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
            int padLength = 39 - strlen(tagBuffer);
            sprintf(tagBuffer, "%*.*s\n", padLength, padLength, " ");
            ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
        }
        ofsBallyBinFile->close();
    }
    tagQue.pop_front();
    close();
}

void BallyBinFile::putNameTag(char *name)
{
    char tagBuffer[512];

    checkParentTag("NAME","BALLY");
    if (createBinFile)
    {
        if (fileFormat == 1)
        {
            sprintf(tagBuffer,"<NAME>%s</NAME>",name);
        }
        else
        {
            sprintf(tagBuffer,"\t<NAME>%s</NAME>\n",name);
        }
        ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
    }
}

void BallyBinFile::putLoadTag()
{
    int tagLength = 0;

    checkParentTag("LOAD","BALLY");
    tagQue.push_front(Tag("LOAD"));
    if (createBinFile)
    {
        if (currLoad->getLoadFormat() == 3)
        {
            unsigned char tagBuffer[80];
            strToBuffer(tagBuffer, &tagLength, "<LOAD format=3>");
            putByte(tagBuffer, &tagLength, ++loadCount);
            ofsBallyBinFile->write(tagBuffer, tagLength);
        } else if (currLoad->getLoadFormat() == 2 || currLoad->getLoadFormat() == 1)
        {
            char tagBuffer[160];
            char gapAttribute[80];

            double header, footer, gap;
            header = currLoad->getHeaderLength();
            footer = currLoad->getFooterLength();
            gap = currLoad->getGapLength();
            if (gap == 0)
                strcpy(gapAttribute, "");
            else
                sprintf(gapAttribute, " gapLength=%f", gap);
            sprintf(tagBuffer, "\t<LOAD format=%d loadNumber=%d headerLength=%f footerLength=%f%s checkSum=0x%2.2x>\n",
                currLoad->getLoadFormat(), ++loadCount, header, footer, gapAttribute, currLoad->getCheckSum());
            tagLength = strlen(tagBuffer);
            ofsBallyBinFile->write(tagBuffer, tagLength);
        }
    }
}

void BallyBinFile::closeLoadTag(void)
{
    char tagBuffer[80];

    verifyTagIsOpen("LOAD");
    if (createBinFile)
    {
        if (fileFormat == 1)
        {
            sprintf(tagBuffer,"</LOAD>");
        }
        else
        {
            sprintf(tagBuffer, "\t</LOAD>\n");
        }
        ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
    }
    tagQue.pop_front();
}

void BallyBinFile::putHeaderTag(double headerLength)
{
    int tagLength = 0;

    checkParentTag("HEADER","LOAD");

    if (createBinFile)
    {
        if (fileFormat == 1)
        {
            unsigned char tagBuffer[80];
            strToBuffer(tagBuffer, &tagLength, "<HEADER>");
            putByte(tagBuffer, &tagLength, floor(headerLength));
            putByte(tagBuffer, &tagLength, floor((headerLength - floor(headerLength))*100));
            strToBuffer(tagBuffer, &tagLength, "</HEADER>");
            ofsBallyBinFile->write(tagBuffer, tagLength);
        } else
        {
            char tagBuffer[80];
            sprintf(tagBuffer, "\t\t<HEADER headerLength=%f></HEADER>\n", headerLength);
            ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
        }
    }
}

void BallyBinFile::putFooterTag(double footerLength)
{
    int tagLength = 0;

    checkParentTag("FOOTER","LOAD");

    if (createBinFile)
    {
        if (fileFormat == 1)
        {
            unsigned char tagBuffer[80];
            strToBuffer(tagBuffer, &tagLength, "<FOOTER>");
            putByte(tagBuffer, &tagLength, floor(footerLength));
            putByte(tagBuffer, &tagLength, floor((footerLength - floor(footerLength))*100));
            strToBuffer(tagBuffer, &tagLength, "</FOOTER>");
            ofsBallyBinFile->write(tagBuffer, tagLength);
        }
        else
        {
            char tagBuffer[80];
            sprintf(tagBuffer, "\t\t<FOOTER footerLength=%f></FOOTER>\n", footerLength);
            ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
            ofsBallyBinFile->flush();
        }
    }
}

void BallyBinFile::putGapTag(double gapLength)
{
    int tagLength = 0;

    if (gapLength >= 0.000001)
    {
        checkParentTag("GAP","LOAD");

        if (createBinFile)
        {
            if (fileFormat == 1)
            {
                unsigned char tagBuffer[80];
                strToBuffer(tagBuffer, &tagLength, "<GAP>");
                putByte(tagBuffer, &tagLength, floor(gapLength));
                putByte(tagBuffer, &tagLength, floor((gapLength - floor(gapLength))*100));
                strToBuffer(tagBuffer, &tagLength, "</GAP>");
                ofsBallyBinFile->write(tagBuffer, tagLength);
            }
            else
            {
                char tagBuffer[80];
                sprintf(tagBuffer, "\t\t<GAP gapLength=%0.6f></GAP>\n", gapLength);
                ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
            }
        }
    }
}

void BallyBinFile::putHexValue(unsigned char byteValue)
{
    char tagBuffer[2];
    sprintf(tagBuffer, "%2.2x ", byteValue);
    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
}

void BallyBinFile::putDataChars(unsigned char *dataBuffer, long dataPosition, int charCount)
{
    char tagBuffer[80];
    int k;
    sprintf(tagBuffer, "<CHAR>", charCount);
    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
    for (k=charCount-1; k>=0; k--)
    {
        char outChar = dataBuffer[dataPosition-k];
        if (outChar >= 32 && outChar <= 126)
        {
            sprintf(tagBuffer, "%c",outChar);
        }
        else
        {
            sprintf(tagBuffer, ".");
        }
        ofsBallyBinFile->write(tagBuffer, 1);
    }
    sprintf(tagBuffer, "</CHAR>");
    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
}

void BallyBinFile::putDataTag(unsigned char *dataBuffer, long dataLength)
{
    int tagLength = 0, dLength;

    checkParentTag("DATA","LOAD");

    if (createBinFile)
    {
        dLength = (dataLength%2 == 0)?dataLength:dataLength+1;
        if (fileFormat == 1)
        {
            unsigned char tagBuffer[80];
            strToBuffer(tagBuffer, &tagLength, "<DATA>");
            putWord(tagBuffer, &tagLength, dataLength);
            ofsBallyBinFile->write(tagBuffer, tagLength);
            tagLength = 0;
            ofsBallyBinFile->write(dataBuffer, dLength);
            strToBuffer(tagBuffer, &tagLength, "</DATA>");
            ofsBallyBinFile->write(tagBuffer, tagLength);
        }
        else
        {
            char tagBuffer[8000];
            long i;
            int j, k, l;
            sprintf(tagBuffer, "\t\t<DATA length=%d>\n", dataLength);
            ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
            basicPointer = 0;
            if (basicFlag)
            {
                long lineNumber;
                char basicLine[8000];

				while (getNextBasicLine(currLoad, &lineNumber, basicLine));
				basicPointer = 0;
				if (basicEndByte != 0)
	                sprintf(tagBuffer, "\t\t\t<BASIC endByte=%d>\n",basicEndByte);
				else
	                sprintf(tagBuffer, "\t\t\t<BASIC>\n");
                ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
                while (getNextBasicLine(currLoad, &lineNumber, basicLine))
                {
                    sprintf(tagBuffer, "\t\t\t\t<LINE number=%4d>%s</LINE>\n", lineNumber, basicLine);
					int mmm=strlen(tagBuffer);
                    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
                    // Remember to put bytes 00 ff at end of basic tag!!!
                }
                sprintf(tagBuffer, "\t\t\t</BASIC>\n");
                ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
            }
            j=0;
            for (i=basicPointer; i < dLength; i++)
            {
                int zeroPointer = i, zeroCount = 0;
                if (j == 0)
                {
                    // Count the number of zeros in a row
                    while (!dataBuffer[zeroPointer] && zeroPointer < dLength)
                    {
                        zeroPointer++;
                        zeroCount++;
                    }
                    // If there are more than 16 zeros, skip them
                    if (zeroCount > 16 || zeroPointer >= dLength)
                    {
                        i = zeroPointer;
                        if (i >= dLength)
                        {
                            i = dLength;
                            break;
                        }
                    }
                    if (dLength - i >= 16)
                        l=16;
                    else
                        l=dLength - i;
                    sprintf(tagBuffer, "\t\t\t<HEX address=%4.4x>", i);
                    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
                }
                putHexValue(dataBuffer[i]);
                j++;
                if (j == 16)
                {
                    putDataChars(dataBuffer, i, j);
                    j=0;
                    sprintf(tagBuffer, "</HEX>\n");
                    ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
                }
            }
            if (j != 0)
            {
                for (k=j; k<16; k++)
                {
                    ofsBallyBinFile->write("   ", 3);
                }
                putDataChars(dataBuffer, i-1, j);
                sprintf(tagBuffer, "</HEX>\n");
                ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
            }
            sprintf(tagBuffer, "\t\t</DATA>\n");
            ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
        }
    }
}

void BallyBinFile::closeDataTag(void)
{
    char tagBuffer[80];

    verifyTagIsOpen("DATA");
    if (createBinFile)
    {
        if (fileFormat == 1)
        {
            sprintf(tagBuffer,"</DATA>");
        }
        else
        {
            sprintf(tagBuffer,"/t/t</DATA>/n");
        }
        ofsBallyBinFile->write(tagBuffer, strlen(tagBuffer));
    }
    tagQue.pop_front();
}

void BallyBinFile::putLoad(int pLoadFormat)
{
    currLoad = &(loadQue.front());
    currLoad->setLoadFormat(pLoadFormat);
    putLoadTag();
    if (fileFormat == 1)
        putHeaderTag(currLoad->getHeaderLength());
    putDataTag(currLoad->dataBuffer, currLoad->getByteCount());
    if (fileFormat == 1)
    {
        putFooterTag(currLoad->getFooterLength());
        putGapTag(currLoad->getGapLength());
    }
	closeLoadTag();
}

void BallyBinFile::putByte(unsigned char *buffer, int *bufferPosition, int byteVal)
{
    buffer[(*bufferPosition)++] = byteVal;
}

void BallyBinFile::putWord(unsigned char *buffer, int *bufferPosition, long inVal)
{
    WordVal::putWord(buffer, *bufferPosition, inVal);
    (*bufferPosition) += 2;
}

void BallyBinFile::strToBuffer(unsigned char *buffer, int *bufferPosition, char *str)
{
    WordVal::strToBuffer(buffer, *bufferPosition, str);
    (*bufferPosition) += strlen(str);
}

void BallyBinFile::convertWord(unsigned char inByteOne, unsigned char inByteTwo,
                               unsigned char *outByteOne, unsigned char *outByteTwo)
{
	/*
       inByteOne  = A7 - A0
	   inByteTwo  = B7 - B0
	   outByteOne = X7 - X0
	   outByteTwo = Y7 - Y0

	   X0 = B7	Y0 = B6 | A0 = Y7  B0 = Y6
       X1 = A7	Y1 = A6 | A1 = X7  B1 = X6
	   X2 = B5	Y2 = B4 | A2 = Y5  B2 = Y4
	   X3 = A5	Y3 = A4 | A3 = X5  B3 = X4
	   X4 = B3	Y4 = B2 | A4 = Y3  B4 = Y2
	   X5 = A3	Y5 = A2 | A5 = X3  B5 = X2
	   X6 = B1	Y6 = B0 | A6 = Y1  B6 = Y0
	   X7 = A1	Y7 = A0 | A7 = X1  B7 = X0
    */

    *outByteOne =
        ((inByteOne & 0x02) > 0)*128 + ((inByteTwo & 0x02) > 0)*64 +
        ((inByteOne & 0x08) > 0)*32  + ((inByteTwo & 0x08) > 0)*16 +
        ((inByteOne & 0x20) > 0)*8   + ((inByteTwo & 0x20) > 0)*4  +
        ((inByteOne & 0x80) > 0)*2   + ((inByteTwo & 0x80) > 0);

    *outByteTwo =
        ((inByteOne & 0x01) > 0)*128 + ((inByteTwo & 0x01) > 0)*64 +
        ((inByteOne & 0x04) > 0)*32  + ((inByteTwo & 0x04) > 0)*16 +
        ((inByteOne & 0x10) > 0)*8   + ((inByteTwo & 0x10) > 0)*4  +
        ((inByteOne & 0x40) > 0)*2   + ((inByteTwo & 0x40) > 0);
}

void BallyBinFile::cleanFileName(char *fileName, char *extension)
{
	if ((strcmpi(fileName + (strlen(fileName) - 4), extension)) == 0)
	{
		if (strlen(fileName) == 4)
			throw Error::badOutputFileName();
		else
			fileName[strlen(fileName) - 4] = '\0';
			
	}
}

void BallyBinFile::setBasicFlag()
{
    basicFlag = 1;
}

void BallyBinFile::clearBasicFlag()
{
    basicFlag = 0;
}

int  BallyBinFile::getBasicFlag()
{
    return basicFlag;
}

int BallyBinFile::getNextBasicLine(TapeLoad *currLoad, long *lineNumber, char *basicLine)
{
    unsigned char currentValue;
    bool firstToken = true;
    strcpy(basicLine, "");
	int lineLength = 0;
    unsigned char *loadBuffer = currLoad->dataBuffer;

	if (loadBuffer[basicPointer+1] == 0xff)
	{
		*lineNumber = -1;
		basicEndByte = loadBuffer[basicPointer];
		basicPointer += 2;
		return 0;
	}

    *lineNumber = WordVal::wordVal(loadBuffer+(basicPointer));

    basicPointer += 2;
/*
    if (*lineNumber == 0xff00 || *lineNumber >= 65447)
    {
        *lineNumber = -1;
        return 0;
    }
*/
    while ((currentValue = loadBuffer[basicPointer++]) != 0x0d)
    {
        switch (currentValue)
        {
        case '"':
            strcat(basicLine, "\"");
            while ((currentValue = loadBuffer[basicPointer++]) != '"')
            {
                if (currentValue == 0x0d)
                    break;
                strcat(basicLine, characterTokens[currentValue]);
            }
            break;
        }
        if (currentValue == 0x0d)
            break;
		lineLength += strlen(basicTokens[currentValue]);
		if (lineLength > 255)
		{
			strcat(basicLine, "<BR>\n\t\t\t\t</BR>");
			lineLength = strlen(basicTokens[currentValue]);
		}
        strcat(basicLine, basicTokens[currentValue]);
    }
    return 1;
}

int BallyBinFile::findBasicToken(char *token, char tokenList[256][40])
{
    int i, tokenCount = 2, tokenFound = 0;
    tokenCount = 0;
    for (i=0; i<256; i++)
    {
        if (!strncmp(token, tokenList[i], strlen(token)))
        {
            tokenCount++;
            tokenFound = i;
        }
    }
    if (tokenCount == 1)
    {
        if (strcmp(token, tokenList[tokenFound]))
            tokenCount = 2;
    }
    if (tokenCount > 1)
        return -1;
    else
        return tokenFound;
}

int BallyBinFile::encodeBasicLine(unsigned char *dataBuffer, long *dataIndex, long lineNumber, char *basicLine)
{
    char tempBuffer[256];
    bool firstToken = true;
    unsigned char lineNumberWord[2];
    int lineLength = strlen(basicLine);
    int lineIndex = 0;
    int bufferIndex = 0;
    int i, j, inQuotes = 0;
/*  for debugging a particular line
	if (lineNumber == 310)
		printf("abc");
*/
    WordVal::decToWord(lineNumber, lineNumberWord);
    dataBuffer[(*dataIndex)++] = lineNumberWord[0];
    dataBuffer[(*dataIndex)++] = lineNumberWord[1];

    memset(tempBuffer, 0, 256);
	int multipleTokensFound = 0;
    for (i=lineIndex; i<lineLength; i++)
    {
        int basicToken;
        // add the next character from the basic line to tempBuffer
        tempBuffer[bufferIndex++] = basicLine[i];
        if (!strcmp(tempBuffer, "\""))
        {
            if (inQuotes)
                inQuotes = 0;
            else
                inQuotes = 1;
        }

        if (inQuotes)
            basicToken = findBasicToken(tempBuffer, characterTokens);
        else
            basicToken = findBasicToken(tempBuffer, basicTokens);

		// We've hit the end of the line and still match multiple tokens
		// just dump the rest of the line
		if (basicToken < 0 && i == lineLength-1)
		{
			basicToken = 0;
		}
		if (basicToken < 0 )
			multipleTokensFound = 1;
		else if (basicToken == 0)
        { // not a token, just dump it in the buffer
            int tokenLength = strlen(tempBuffer);
			if (multipleTokensFound) // Found more than 1 tokens before, but none this time
			{
				// Copy all but the last character to the basic string
				tokenLength -= 1;
			}
            for (j=0; j<tokenLength; j++)
            {
                dataBuffer[(*dataIndex)++] = tempBuffer[j];
            }
			if (multipleTokensFound)
			{
				i--; // Move pointer to previous character
				multipleTokensFound = 0;
			}
	        memset(tempBuffer, 0, 256);
	        bufferIndex = 0;
        } else if (basicToken > 0) {
            dataBuffer[(*dataIndex)++] = basicToken;
            memset(tempBuffer, 0, 256);
            bufferIndex = 0;
        }
    }
    dataBuffer[(*dataIndex)++] = 0x0d;
    return 0;
}

void BallyBinFile::setFileFormat(int pFileFormat)
{
    fileFormat = pFileFormat;
}

int  BallyBinFile::getFileFormat()
{
    return fileFormat;
}

void BallyBinFile::getTapeName(char *pLoadName)
{
    strcpy(pLoadName, loadName);
}

void BallyBinFile::setTapeName(char *pLoadName)
{
    strcpy(loadName, pLoadName);
}

void BallyBinFile::getScreenBmpName(char *pScreenBmpName)
{
	strcpy(pScreenBmpName, screenBmpName);
}

void BallyBinFile::setScreenBmpName(char *pScreenBmpName)
{
	strcpy(screenBmpName, pScreenBmpName);
}

Tag::Tag()
{
};

Tag::Tag(std::string const &str): name(str)
{
    Tag();
}

Tag::Tag(std::string const &str, double pFilePosition): name(str), filePosition(pFilePosition)
{
    Tag();
}

TapeLoad::TapeLoad()
{
    memset(dataBuffer, 0, BUFFERSIZE);
    maxBytePointer = 0;
    setHeaderLength(1);
    setFooterLength(0.25);
    setGapLength(0);
};

void TapeLoad::toString()
{
    printf("Load Number         = %d\n", getLoadNumber());
    printf("Header Length       = %f\n", getHeaderLength());
    printf("Data Byte Count     = %d\n", getByteCount());
    printf("Footer Length       = %f\n", getFooterLength());
    printf("Gap Length          = %f\n", getGapLength());
    printf("Read Checksum       = %x\n", getCheckSum());
    printf("Calculated Checksum = %x\n", calculateCheckSum());
}

int TapeLoad::getLoadNumber()
{
    return loadNumber;
}

void TapeLoad::setLoadNumber(int pLoadNumber)
{
    loadNumber = pLoadNumber;
}

void TapeLoad::setMaxBytePointer(long pBytePointer)
{
    if (pBytePointer > maxBytePointer)
        maxBytePointer = pBytePointer;
}

long TapeLoad::getMaxBytePointer()
{
    return maxBytePointer;
}

void TapeLoad::resetMaxBytePointer()
{
    maxBytePointer = 0;
}

void TapeLoad::setHeaderLength(double pHeaderLength)
{
    headerLength = pHeaderLength;
}

double TapeLoad::getHeaderLength()
{
    return headerLength;
}

void TapeLoad::setGapLength(double pGapLength)
{
    gapLength = pGapLength;
}

double TapeLoad::getGapLength()
{
    return gapLength;
}

void TapeLoad::setByteCount(long pByteCount)
{
    byteCount = pByteCount;
}

long TapeLoad::getByteCount()
{
    return byteCount;
}

void TapeLoad::setFooterLength(double pFooterLength)
{
    footerLength = pFooterLength;
}

double TapeLoad::getFooterLength()
{
    return footerLength;
}

void TapeLoad::setLoadFormat(int pLoadFormat)
{
    loadFormat = pLoadFormat;
}

int TapeLoad::getLoadFormat()
{
    return loadFormat;
}

void TapeLoad::setCheckSum(unsigned char pCheckSum)
{
    checkSum = pCheckSum;
}

unsigned char TapeLoad::getCheckSum()
{
    return checkSum;
}

unsigned char TapeLoad::calculateCheckSum()
{
    unsigned char lCheckSum = 0xa5;
    if (getLoadFormat() == 1)
    {
        for (int i=0; i < getByteCount(); i++)
        {
            lCheckSum += dataBuffer[i];
        }
    } else {
        unsigned char outByteOne, outByteTwo;
		double bc = getByteCount() / 2.0;
        long loadOffset = ceil(bc);
        for (int i=0; i < loadOffset; i++)
        {
//            BallyBinFile::convertWord(dataBuffer[i], dataBuffer[i+loadOffset+1], &outByteOne, &outByteTwo);
            BallyBinFile::convertWord(dataBuffer[i], dataBuffer[i+0x1000], &outByteOne, &outByteTwo);
            lCheckSum += WordVal::revByte(outByteOne);
            if (i+loadOffset < getByteCount())
            {
                lCheckSum += WordVal::revByte(outByteTwo);
            }
        }
    }
//	dump();
    return 0xff - lCheckSum;
}

void TapeLoad::dump()
{
	int i, j;
	for (i=0; i<0x200; i++)
	{
		printf("%4.4x - ", i*0x10);
		for (j=0; j<0x10; j++)
		{
			printf("%2.2x ", dataBuffer[i*0x10+j]);
		}
		printf("\n");
	}
}

void TapeLoad::convertLoadFormat(int pNewLoadFormat)
{
    unsigned char tempDataBuffer[BUFFERSIZE];

    if (pNewLoadFormat == getLoadFormat())
        return;
    if (pNewLoadFormat == 1 && getLoadFormat() == 3)
    {
        setLoadFormat(1);
        long loadOffset = ceil(getByteCount() / 2) + 1;
        int i;
        for (i = 0; i < loadOffset; i++)
        {
            BallyBinFile::convertWord(dataBuffer[i], dataBuffer[i+loadOffset], &(tempDataBuffer[i*2]), &(tempDataBuffer[(i*2)+1]));
        }
        setByteCount(getByteCount()-1);
        for (i = 0; i < getByteCount(); i++)
        {
            dataBuffer[i] = WordVal::revByte(tempDataBuffer[i]);
        }
        setCheckSum(calculateCheckSum());
    } else if (pNewLoadFormat == 3 && getLoadFormat() == 1)
    {
        setLoadFormat(3);
        for (int i = 0; i < getByteCount(); i++)
        {
            tempDataBuffer[i] = WordVal::revByte(dataBuffer[i]);
        }
        tempDataBuffer[getByteCount()] = WordVal::revByte(getCheckSum());
        setByteCount(getByteCount()+1);
        tempDataBuffer[getByteCount()] = 0xff;
        long loadOffset = ceil(getByteCount() / 2) + 1;
        for (i = 0; i < loadOffset; i++)
        {
            if (i*2+1 == getByteCount())
                printf("");
            BallyBinFile::convertWord(tempDataBuffer[i*2], tempDataBuffer[(i*2)+1], &(dataBuffer[i]), &(dataBuffer[i+loadOffset]));
        }
        setCheckSum(calculateCheckSum());
    } else
    {
        cerr << "Invalid format conversion from " << getLoadFormat() << " to " << pNewLoadFormat;
        exit(1);
    }
}
