#ifndef BALLY_BIN_FILE_DEFINED
#define BALLY_BIN_FILE_DEFINED
#include <cstdio>
#include <cstdlib>
#include <string>
#include <fstream>
#include <iostream>
#include <math.h>
#include <deque>
#include "BallyWavOut.h"

#ifdef USE_IPL98
	#include <image.h>
#endif

#define  BUFFERSIZE 0x2000
#define  PIXEL  219

class TapeLoad
{
    friend class BallyBinFile;

public:
    TapeLoad();
    unsigned char getCheckSum();
    void          setCheckSum(unsigned char);
    double getHeaderLength();
    void   setHeaderLength(double);
    double getFooterLength();
    void   setFooterLength(double);
    double getGapLength();
    void   setGapLength(double);
    int    getLoadFormat();
    void   setLoadFormat(int);
    long   getByteCount();
    void   setByteCount(long);
    long   getMaxBytePointer();
    void   setMaxBytePointer(long);
    int    getLoadNumber();
    void   setLoadNumber(int);
	void   dump();

    unsigned char calculateCheckSum();
    void   resetMaxBytePointer();
    void   convertLoadFormat(int);
    void   toString();

    operator TapeLoad const &() const
    {
        return (*this);
    }

    bool operator<(TapeLoad const &right) const
    {
        return (loadNumber < right.loadNumber);
    }

    bool operator<=(TapeLoad const &right) const
    {
        return (loadNumber <= right.loadNumber);
    }

    bool operator>(TapeLoad const &right) const
    {
        return (loadNumber > right.loadNumber);
    }

    bool operator>=(TapeLoad const &right) const
    {
        return (loadNumber >= right.loadNumber);
    }

    bool operator==(TapeLoad const &right) const
    {
        return (loadNumber == right.loadNumber);
    }

    bool operator!=(TapeLoad const &right) const
    {
        return (loadNumber != right.loadNumber);
    }

    unsigned char dataBuffer[BUFFERSIZE];

protected:
    int loadNumber;

private:
    int    loadFormat;
    long   byteCount;
    double gapLength;
    double headerLength,
           footerLength;
    unsigned char checkSum;
    long maxBytePointer;
};

class Tag
{
    friend class BallyBinFile;
    public:

    Tag();

    Tag(std::string const &str);

    Tag(std::string const &str, double const pFilePosition);

    operator std::string const &() const
    {
        return (name);
    }
    operator double const &() const
    {
        return (filePosition);
    }
    operator Tag const &() const
    {
        return (*this);
    }
    bool operator<(Tag const &right) const
    {
        return (name < right.name);
    }
    bool operator<=(Tag const &right) const
    {
        return (name <= right.name);
    }
    bool operator>(Tag const &right) const
    {
        return (name > right.name);
    }
    bool operator>=(Tag const &right) const
    {
        return (name >= right.name);
    }
    bool operator==(Tag const &right) const
    {
        return (name == right.name);
    }
    bool operator!=(Tag const &right) const
    {
        return (name != right.name);
    }

protected:
    std::string name;
    double filePosition;
};

class BallyBinFile {
public:
	BallyBinFile();
	~BallyBinFile();

	int open(char *, char *, int);
    void close();

    void writeWavFile(char *, long pSampleRate = 22050, int pBitsPerSample = 8, int pChannels = 1);
	void putLoad(int);
    void setBasicFlag();
    void clearBasicFlag();
    int  getBasicFlag();
    int  findBasicToken(char *token, char [256][40]);
    int  decodeWavFile(char *inputName);
    void writeBinFile(char *binName);
    int  encodeBasicLine(unsigned char *dataBuffer, long *dataIndex, long lineNumber, char *basicLine);
	int  getNextBasicLine(TapeLoad *, long *, char *basicLine);
    void getTapeName(char *);
    void setTapeName(char *);
    void putHexValue(unsigned char);
    void putDataChars(unsigned char *dataBuffer, long dataPosition, int charCount);
    int  getShowScreen();
    void setShowScreen(int);
	void getScreenBmpName(char *);
	void setScreenBmpName(char *);
    int  getLastBasicLoad();
    void setLastBasicLoad(int);
    int  getFirstDataLoad();
    void setFirstDataLoad(int);
    int  getMinHalfWavLength();
    void setMinHalfWavLength(int);
    int  getVerbosity();
    void setVerbosity(int);
    void putBallyTag(int pLoadCount);

    static void convertWord(unsigned char inByteOne,   unsigned char inByteTwo,
                            unsigned char *outByteOne, unsigned char *outByteTwo);
    static void cleanFileName(char *fileName, char *extension);
    static void putByte(unsigned char *buffer, int *bufferPosition, int byteVal);
    static void putWord(unsigned char *buffer, int *bufferPosition, long inVal);
    static void strToBuffer(unsigned char *buffer, int *bufferPosition, char *str);
	int processTag(char *);

    void setFileFormat(int pFileFormat);
    int  getFileFormat();

    void getAttributeVal(char *tag, char *pAttribute, char *attributeVal);

    bool createBinFile;

	double pHeaderLength, pFooterLength, pGapLength;

    std::deque<TapeLoad> loadQue;

private:
	ifstream ifsBallyBinFile;
    ofstream *ofsBallyBinFile;

    std::deque<Tag> tagQue;

    BallyWavOut wavFile;
    TapeLoad *currLoad;

	long sampleRate,
	     basicPointer;

    int basicFlag,
	    bitsPerSample,
        channels,
	    currentLoad,
	    loadNumber,
        verbosity,
        multiLoadNumber,
        fileFormat,
        showScreen,
        lastBasicLoad,
        minHalfWavLength,
        firstDataLoad,
        loadToProcess,
	    loadCount,
        hexLength,
		basicEndByte;

	double fileSize,
           ballyTagPos,
           headerTagPos,
           footerTagPos,
           loadTagPos,
           fileLength,
           checkSum,
           loadLength;

	char tapeName[256],
         loadName[256],
         binName[256],
         basicTokens[256][40],
         characterTokens[256][40],
		 screenBmpName[256];

    unsigned char byte,
	              word[2],
	              longWord[4],
                  *loadBuffer;

    int processOpenTagData(char *token, char *tag, char *fullTag);
    int processBallyTagData(char *token, char *tag, char *fullTag);
    int processLoadTagData(char *token, char *tag, char *fullTag);
    int processDataTagData(char *token, char *tag, char *fullTag);
    int processHexTag(char *token, char *tag, char *fullTag);
    int processBasicTag(char *token, char *tag, char *fullTag);
    int processLineTag(char *token, char *tag, char *fullTag);
    int processBrTag();
    void putNameTag(char *);
    void putLoadTag(void);
    void putHeaderTag(double);
    void putFooterTag(double);
    void putGapTag(double);
    void putDataTag(unsigned char *dataBuffer, long dataLength);

    void checkParentTag(std::string, std::string);
    void verifyTagIsOpen(std::string tagName);
	void getTagValue(char *endTag, char *tagValue);
    void closeDataTag(void);
    void closeLoadTag(void);
    void closeBallyTag(void);

	struct basicTokenType
	{
		unsigned char value;
		char mnemonic[40];
        int inQuotes;
	};

};
#endif