#include <cstdlib>
#include <string>
#include <iostream>
#include "Error.h"
#include "ballybin.h"
#include "ballyBinFile.h"
//#include "BallyWavIn.h"

#ifdef MSDOS
    #include <wtypes.h>
    #include <wincon.h>
#endif

void displayProgramInfo()
{
    cout << "BallyBin v2.1\n\n";
	cout << "Released January 10, 2005\n";
	cout << "Written by: Bob Colbert\n\n";
    cout.flush();
}

void getParmString(int *argc, char ***argv, char *returnString, int maxLength)
{
	if (strlen((*argv)[0]) == 1)
	{
		if (*argc == 1)
			throw Error::badParmValue();

		if ((*argv)[1][0] == '-')
			throw Error::badParmValue();

		strncpy(returnString, (*argv)[1], maxLength);
		--(*argc);
		(*++(*argv));
	} else
	{
		strncpy(returnString, **argv + 1, maxLength);
	}
	(**argv)[1] = 0;
}

int main(int argc, char *argv[])
{
	char binName[256]    = "",
         wavName[256]    = "",
         parmString[20]  = "",
         tapeName[256]   = "",
         inputName[256]  = "";

	int c                = 0, // Used for parsing argument flags
        i                = 0, // Generic increment variable
        pFormat          = 0,
        minHalfWavLength = 0,
        lastBasicLoad    = 0,
        verbosity        = 1,
        showScreen       = 0,
        firstDataLoad    = 2,
        loadCount        = 0;

	bool writeWavFile = false;

	BallyBinFile binFile;

    binFile.setFileFormat(2);

	if (argc == 1)
	{
        displayProgramInfo();
		cout << "Usage:\n";
		cout << "   BallyBin [options] (input filename - names with spaces require quotes)\n";
		cout << "\n";
		cout << "Standard options:\n";
        cout << "   [-d(n)]      Decode the first n loads as BASIC.\n";
		cout << "   [-s(name)]   Show/save screen for BASIC loads. ";
#ifdef USE_IPL98
		cout <<                                                   "Bitmap picture will be saved as\n";
		cout << "                (input filename).bmp if no name is provided.\n";
#endif
		cout << "                Names with spaces require quotes.\n";
		cout << "   [-w(name)]   Create a remastered .wav file. The .wav file will be saved as\n";
		cout << "                Remastered_(input filename).wav if no name is provided.\n";
		cout << "                (e.g. game.wav would create Remastered_game.wav)\n";
		cout << "                Names with spaces require quotes.\n";
		cout << "\n";
		cout << "Advanced options:\n";
		cout << "   [-b(name)]   The bml output file name, default is the input file name with\n";
		cout << "                the .bml extension. Names with spaces require quotes.\n";
        cout << "   [-c]         Convert between .bml and .bin formats. If this flag is used then\n";
		cout << "                the input file should be a .bml or .bin.\n";
        cout << "   [-f(format)] File format: 1, (2), or 3.\n";
        cout << "   [-i(load #)] First load to be a data load (default = 2).\n";
		cout << "   [-l(length)] Minimum half-wavelength.\n";
		cout << "   [-n(name)]   Tape name to be stored in .bml file (can not contain '<' character).\n";
        cout << "   [-v(number)] Verbosity of 0,(1) or 2 where 0 is silent and 2 is the most detailed.\n\n";

		return 1;
	}

	try
	{
		// Process command line arguments
		while (--argc > 0 && (*++argv)[0] == '-')
		{
			while (c = *++argv[0])
			{
				switch (c)
				{
					case 'b':
						try
						{
							getParmString(&argc, &argv, binName, 255);
						}
						catch (Error::badParmValue)
						{
							cerr << "Invalid file name for b flag\n";
							throw;
						}
						break;
					case 'c':
						pFormat = 1;
						break;
					case 'd':
						try
						{
							getParmString(&argc, &argv, parmString, 19);
							lastBasicLoad = atoi(parmString);
						}
						catch (Error::badParmValue)
						{
							cerr << "Invalid load count for d flag\n";
							throw;
						}
						break;
					case 'f':
						try
						{
							getParmString(&argc, &argv, parmString, 19);
							binFile.setFileFormat(atoi(parmString));
							if (binFile.getFileFormat() > 2)
							{
								throw Error::badParmValue();
							}
						}
						catch (Error::badParmValue)
						{
							cerr << "Invalid file format for f flag of " << binFile.getFileFormat() << "\n";
							throw;
						}
						break;
					case 'i':
						try
						{
							getParmString(&argc, &argv, parmString, 19);
							firstDataLoad = atoi(parmString);
						}
						catch (Error::badParmValue)
						{
							cerr << "Invalid initial load value for i flag\n";
							throw;
						}
						break;
					case 'l':
						try
						{
							getParmString(&argc, &argv, parmString, 19);
							minHalfWavLength = atoi(parmString);
							if (minHalfWavLength < 2)
								throw Error::badParmValue();
						}
						catch (Error::badParmValue)
						{
							cerr << "Invalid minimum half wave length for l flag\n";
							throw;
						}
						break;
					case 'n':  // name in <NAME> tag
						try
						{
							getParmString(&argc, &argv, tapeName, 255);
							binFile.setTapeName(tapeName);
						}
						catch (Error::badParmValue)
						{
							cerr << "Invalid name value for n flag\n";
							throw;
						}
						break;
					case 's':
						if (argv[0][1] != 0)
						{
							getParmString(&argc, &argv, parmString, 255);
							binFile.setScreenBmpName(parmString);
						}
						showScreen = 1;
						break;
					case 'v':
						try
						{
							getParmString(&argc, &argv, parmString, 19);
							verbosity = atoi(parmString);
						}
						catch (Error::badParmValue)
						{
							cerr << "Invalid verbosity for v flag\n";
							throw;
						}
						break;
					case 'w':
						if (argv[0][1] != 0)
							getParmString(&argc, &argv, wavName, 255);
						writeWavFile = true;
						break;
					default:
						cout << "Invalid option " << (char)c << "\n";
						return 1;
						break;
				}
			}
		}
    }
	catch (Error::badParmValue)
	{
		cerr << "Invalid Arguments, aborting\n";
		exit(1);
	}

    if (verbosity)
    {
        displayProgramInfo();
    }

    // The remaining argument should be the audio file to process or the binary file to convert
    strcpy(inputName, argv[0]);

    // Convert from one format to another
    if (pFormat > 0)
    {
        // Open the BallyBinFile in read mode and read in the binary file
        binFile.open(argv[0], "r", verbosity);
        binFile.close();

        // Switch formats
        if (binFile.getFileFormat() == 1)
            pFormat = 2;
        else if (binFile.getFileFormat() == 2)
            pFormat = 1;
        else
        {
            cerr << "Unknown format, unable to convert\n";
            exit(1);
        }
        binFile.setFileFormat(pFormat);
        binFile.setLastBasicLoad(lastBasicLoad);
        binFile.setFirstDataLoad(firstDataLoad);
        binFile.setVerbosity(verbosity);
        if (!strcmp(binName, ""))
        {
            BallyBinFile::cleanFileName(argv[0], ".bin");
            BallyBinFile::cleanFileName(argv[0], ".bml");
            sprintf(binName, argv[0]);
        }
        binFile.writeBinFile(binName);
        if (verbosity)
        {
            cout << "Done coverting!\n";
        }
        exit(0);
    }

#ifdef MSDOS
    HANDLE     myHandle,
               newHandle;
    COORD      conCoord;
    SMALL_RECT winRect;
    if (showScreen)
    {
        unsigned long charCount;

        conCoord.X = 160;
        conCoord.Y = 40;

        winRect.Top    = 0;
        winRect.Left   = 0;
        winRect.Bottom = conCoord.Y - 1;
        winRect.Right  = conCoord.X - 1;

        newHandle = CreateConsoleScreenBuffer(GENERIC_WRITE, 0, NULL, CONSOLE_TEXTMODE_BUFFER, NULL);
        if (newHandle == INVALID_HANDLE_VALUE)
        {
            cerr << "Can not create console handle\n";
            exit(1);
        }
        SetConsoleTitle("BallyBin");
        myHandle = GetStdHandle(STD_OUTPUT_HANDLE);
        SetConsoleScreenBufferSize(newHandle, conCoord);
        SetConsoleWindowInfo(newHandle, TRUE, &winRect);
        SetConsoleScreenBufferSize(myHandle, conCoord);
        SetConsoleWindowInfo(myHandle, TRUE, &winRect);
        SetConsoleTextAttribute(newHandle, BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE);
        FillConsoleOutputAttribute(newHandle, BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE, conCoord.X * conCoord.Y, conCoord, &charCount);
    }
#endif

    binFile.setShowScreen(showScreen);
    binFile.setLastBasicLoad(lastBasicLoad);
    binFile.setMinHalfWavLength(minHalfWavLength);
    binFile.setFirstDataLoad(firstDataLoad);
    binFile.setVerbosity(verbosity);
	char tmpName[256], scrName[256];
	strcpy(tmpName, inputName);
    BallyBinFile::cleanFileName(inputName, ".wav");
	if (!strlen(binName) && strlen(inputName))
	{
		strcpy(binName, inputName);
	}
    BallyBinFile::cleanFileName(binName, ".bin");
    BallyBinFile::cleanFileName(binName, ".bml");
	binFile.getScreenBmpName(scrName);
	if (!strcmp(scrName, ""))
	{
		sprintf(scrName, "%s.bmp", binName);
		binFile.setScreenBmpName(scrName);
	}
    if (binFile.decodeWavFile(tmpName) == 0)
    {
        if (verbosity)
        {
            if (binFile.getFileFormat() == 1)
            {
                cout << "Creating bin file " << binName << ".bin \n";
            }
            else
            {
                cout << "Creating bml file " << binName << ".bml \n";
            }
        }
        binFile.writeBinFile(binName);
        if (verbosity)
        {
	   	    cout << "Successfully converted audio file " << inputName << ".wav!\n\n";
        }
        if (writeWavFile)
        {
            BallyBinFile inFile;

            BallyBinFile::cleanFileName(wavName, ".wav");
            if (!strcmp(wavName, inputName) || !strlen(wavName))
            {
                sprintf(wavName, "Remastered_%s", inputName);
            }

            char inFileName[160];
            sprintf(inFileName, "%s.bml", binName);
            inFile.open(inFileName,"r",verbosity);
            inFile.writeWavFile(wavName, 22050, 8, 1);
            if (verbosity)
            {
    	   	    cout << "Successfully created audio file " << wavName << ".wav!\n\n";
            }
        }
    }

#ifdef MSDOS
    if (showScreen)
    {
        SetConsoleActiveScreenBuffer(myHandle);
    }
#endif
	return 0;
}
