#include "scWavFile.h"

ScDataFile::ScDataFile()
{
    for(int page=0;page<0x100;page++)
    {
        for(int byte=0;byte<0x100;byte++)
            buffer[page][byte]=0;
        bankByte[page] = 0;
        checksum[page] = 0;
    }

    fileOpen       = 0;
    headerTime     = 0;
    byteCount      = 0;
    footerTime     = 0;
    startLoByte    = 0;
    startHiByte    = 0;
    controlByte    = 0;
    pageCount      = 0;
    headerChecksum = 0;
    multiLoadByte  = 0;
    speedLoByte    = 0;
    speedHiByte    = 0;
}

void ScDataFile::open(char *pFileName)
{
    char fileName[256];

    if (!fileOpen)
    {
        if (pFileName == 0)
            strcpy(fileName,"d:\\emulate\\atari\\aaa.bin");
        else
            strcpy(fileName,pFileName);

        outFile.open(fileName,ios::out|ios::binary|ios::trunc);
        fileOpen = 1;
    }
}

void ScDataFile::write()
{
    unsigned char outByte;
    for(int page=0; page < 0x20; page++)
    {
        for(int byte=0; byte < 0x100; byte++)
        {
            if (page < pageCount)
                outByte = buffer[page][byte];
            else
                outByte = 0;
            outFile.write(&outByte,1);
        }
    }

    outFile.write(&startLoByte,1);
    outFile.write(&startHiByte,1);
    outFile.write(&controlByte,1);
    outFile.write(&pageCount,1);
    outFile.write(&headerChecksum,1);
    outFile.write(&multiLoadByte,1);
    outFile.write(&speedLoByte,1);
    outFile.write(&speedHiByte,1);
    outByte = 0;
    for(int byte = 0; byte < 8; byte++)
        outFile.write(&outByte,1);
    outFile.write(bankByte,0x30);
    outFile.write(checksum,0xc0);
    outFile.close();
}

ScWavFile::ScWavFile()
{
    fileSynced = false;
    bitsSynced = false;
    clearingToneLength = 0;
    loadNumber = 0;
}

ScWavFile::~ScWavFile(){}

void ScWavFile::open(char *inFile)
{
    WavDataFile::open(inFile);
    sync();
}

unsigned int ScWavFile::readByteHeader(void)
{
    sync();
    byteStream = 0;
    return decodeBits(2);
}

void ScWavFile::sync(void)
{
    unsigned char testByte;
    bool foundClearingTone = 0;

    while (!fileSynced)
    {
        syncLength = syncFile();
        if (foundClearingTone)
            clearingToneLength += syncLength;
        if (alternatingHeader)
            wavLengthMidPoint = sineWaveLengthAvg * 2;
        else
        {
            oneBitLength = floor(sineWaveLengthAvg * 2 + .5);
	        zeroBitLength = floor(sineWaveLengthAvg + .5);
            wavLengthMidPoint = (zeroBitLength + oneBitLength) / 2;
        }
        testByte = readByte();
        if (testByte != 0x55 && testByte !=0xaa)
        {
            if (testByte == 0xff)
                foundClearingTone = 1;
            else
                clearingToneLength = 0;
            fileSynced = false;
            wavLengthAvg.reset(50);
        }
        else
        {
            cout << "File synchronized in " << syncLength << " seconds.\n";
            cout.flush();
            wavLengthMidPoint = sineWaveLengthAvg * 2; //(zeroBitLength + oneBitLength) / 2;
        }
    }
}

long ScWavFile::decodeFooter(void)
{
    long footerLength = 0;
    WavTimer footerTime(this);
    unsigned char footerByte;

    try
    {
        footerByte = readByte();
        while (footerByte == 0x55 || footerByte == 0xd2)
        {
            footerLength += 1;
            showProgress("Footer",footerLength);
            footerByte = readByte();
        }
        throw Error::footerConverted();
    }

    catch(Error::footerConverted)
    {
        if (footerByte != 0x54)
            bitsSynced = false;
        cout << "Decoded " << footerLength << " footer bytes\n";
        cout.flush();
    }

    catch(Error::endOfFile)
    {
        cout << "Decoded " << footerLength << " footer bytes\n";
        cout << "Footer length =" << footerTime.currentTime() << " seconds\n";
        cout << "Input file ended while decoding footer\n";
        load[loadNumber]->footerTime = footerTime.currentTime();
        footerTime.~WavTimer();
        throw Error::conversionComplete();
    }
    cout << "Footer length " << footerTime.currentTime() << "\n";
    load[loadNumber]->footerTime = footerTime.currentTime();
    footerTime.~WavTimer();
    return footerLength;
}

bool ScWavFile::decodeLoad(void)
{
    int loadByteNum = 0;
    unsigned int page;
    ScDataFile *thisLoad = load[loadNumber];
    int byte;
    int badByte = 0;
    unsigned char checksum = 0;
    try
    {
        for (page=0; page<thisLoad->pageCount; page++)
        {
            checksum = 0;
            badByte = -1;
            checksum += (thisLoad->bankByte[page] = readByte());
            checksum += (thisLoad->checksum[page] = readByte());
            for (byte=0; byte<0x100; byte++)
            {
                checksum += (thisLoad->buffer[page][byte] = readByte());
                if (bitError && badByte < 0)
                    badByte = byte;
                showProgress("Data",page * 0x100 + byte);
            }
            if (checksum != 0x55)
            {
                cout << "Bad bank checksum\n";
                if (badByte >=0)
                {
                    cout << "Attempting to fix, bad byte=" << (unsigned int)thisLoad->buffer[page][badByte] << "\n";
                    thisLoad->buffer[page][badByte] += 0x55 - checksum;
                    cout << "                  new value=" << (unsigned int)thisLoad->buffer[page][badByte] << "\n";
                }

                throw Error::badPageChecksum();
            }
        }
        throw Error::loadConverted();
    }

    catch(Error::endOfFile)
    {
        cout << "Input file ended while decoding data\n";
        loadNumber--;
        throw Error::endOfFile();
    }

    catch(Error::loadConverted)
    {
        load[loadNumber]->byteCount = page * 0x100;
        cout << "Decoded " << page * 0x100 << " bytes\n";
    }

    catch(Error::badPageChecksum)
    {
        throw Error::endOfFile();
    }

    return false;
}

int ScWavFile::convert(char *inFile)
{
    int result;
    try
    {
        load[loadNumber] = new ScDataFile;
        open(inFile);
        while (1)
        {
            load[loadNumber]->headerTime = decodeHeader(0x54);
            decodeLoad();
            decodeFooter();
            load[loadNumber]->open(0);
            load[loadNumber]->write();
            load[++loadNumber] = new ScDataFile;
        }
    }

    catch(Error::conversionComplete)
    {
        cout << "Conversion completed successfuly.\n";
        result = 0;
    }

    catch(Error::endOfFile)
    {
        cout << "Conversion was unsuccessful.\n";
        result = 1;
    }
    cout << "clearing tone = " << clearingToneLength << "\n";

    for (int i = 0; i <= loadNumber; i++)
    {
        load[loadNumber]->open(0);
        load[loadNumber]->write();
        if (i>0 && load[i]->headerTime < 0.25)
        {
            load[i-1]->footerTime = (load[i-1]->footerTime + load[i]->headerTime)/2;
            load[i]->headerTime = load[i-1]->footerTime;
        }
    }
    for (i = 0; i <= loadNumber; i++)
    {
        cout << "Load " << i
             << " header="<< load[i]->headerTime
             << " bytes=" << load[i]->byteCount
             << " footer=" << load[i]->footerTime
             << "\n";
    }
    return result;
}

double ScWavFile::decodeHeader(unsigned int headerEndVal)
{
    float headerByteLength = 0;
    WavTimer headerTime(this);
    unsigned char checksum;
    ScDataFile *thisLoad = load[loadNumber];

    try
    {
        if (!bitsSynced)
        {
	        while (byteStream != headerEndVal)
	        {
                decodeBits(1);
                headerByteLength += (float)(0.125);
                showProgress("Header",(int)headerByteLength);
	        }
	        zeroBitLength = lastZeroBitLength;
            oneBitLength = lastOneBitLength;
	        cout << "zeroBitLength = " << zeroBitLength << "\n";
            cout << "oneBitLength  = " << oneBitLength << "\n";
            cout << "Header length = " << headerTime.currentTime() << " seconds.\n";
            cout.flush();
            headerTime.~WavTimer();
            bitsSynced = true;
        }
        checksum = 0x00;
        checksum += (thisLoad->startLoByte = readByte());
        checksum += (thisLoad->startHiByte = readByte());
        checksum += (thisLoad->controlByte = readByte());
        checksum += (thisLoad->pageCount   = readByte());
        checksum += (thisLoad->headerChecksum = readByte());
        checksum += (thisLoad->multiLoadByte  = readByte());
        checksum += (thisLoad->speedLoByte    = readByte());
        checksum += (thisLoad->speedHiByte    = readByte());
        cout.setf(ios::hex,ios::basefield);
        cout << "Start address = "<< (unsigned int)thisLoad->startHiByte << (unsigned int)thisLoad->startLoByte<<"\n";
        cout << "Control byte = "<<(unsigned int)thisLoad->controlByte <<"\n";
        cout.setf(ios::dec,ios::basefield);
        cout << "Page count = "<<(unsigned int)thisLoad->pageCount <<"\n";
        cout.setf(ios::hex,ios::basefield);
        cout << "Checksum = "<<(unsigned int)thisLoad->headerChecksum << " is ";
        cout.setf(ios::dec,ios::basefield);
        if (checksum == 0x55)
            cout << "valid\n";
        else
        {
            cout << "invalid\n";
            throw Error::badHeaderChecksum();
        }
    }

    catch(Error::endOfFile)
    {
        cout << "Input file ended while decoding header\n";
        loadNumber--;
        throw Error::conversionComplete();
    }

    catch(Error::badHeaderChecksum)
    {
        throw Error::endOfFile();
    }

    return headerTime.currentTime();
}
