#include <cstdlib>
#include <string>
#include <iostream.h>
#include "ballyBinFile.h"
#include "Error.h"

unsigned char binBuffer[0x6000];
unsigned char loadBuffer[0x1000];

void getParmString(int *argc, char ***argv, char *returnString, int maxLength)
{
	if (strlen((*argv)[0]) == 1)
	{
		strncpy(returnString, (*argv)[1], maxLength);
		--(*argc);
		(*++(*argv));
	} else
	{
		strncpy(returnString, **argv + 1, maxLength);
	}
	(**argv)[1] = 0;
}

int main(int argc, char *argv[]) {
	char binName[256]="", wavName[256]="";
	int binByteNum, loadByteNum, channelCount = 2, bitCount = 8;
	int c, startLoad, endLoad;
	long sampleRate = 22050;
	double pHeaderLength = -1.0,
		   pFooterLength = -1.0,
		   pGapLength    = -1.0;
	char tapeName[256] = "";
	char parmString[20] = "";
    bool splitMultiLoad = false;
	ofstream ofsScBin;
	BallyBinFile inFile, validFile;
	int verbosity = 1;
    binByteNum = 0;
    loadByteNum = 0;

	if (argc == 1)
	{
		cerr << "Usage:\n";
		cerr << "   AstroWav [-b ] ([8], 16) number of bits per sample in .wav file\n";
		cerr << "            [-c ] ([1], 2) number of audio channels\n";
		cerr << "            [-e ] last load to process - default = process all loads\n";
		cout << "            [-f ] footer length in seconds (e.g. 2.5)\n";
		cout << "            [-g ] gap length in seconds\n";
		cout << "            [-h ] header length in seconds\n";
		cout << "            [-k ] ([22050], 44100) sample rate\n";
		cout << "            [-m ] (split multi-loads - default = don't split)\n";
		cout << "            [-s ] first load to process - default = 1\n";
        cout << "            [-v ] verbosity 0-2 where 0 is silent and 2 is the most detailed\n";
        cout << "                  Default = 1\n";
		cout << "            [-w (.wav file name)]\n";
		cerr << "            (input .bin filename)\n";

		return 1;
	}

	while (--argc > 0 && (*++argv)[0] == '-')
		while (c = *++argv[0])
			switch (c)
			{
				case 'b':
					getParmString(&argc, &argv, parmString, 19);
					bitCount = atoi(parmString);
					break;
				case 'c':
					getParmString(&argc, &argv, parmString, 19);
					channelCount = atoi(parmString);
					break;
				case 'e':
					getParmString(&argc, &argv, parmString, 19);
					endLoad = atoi(parmString);
					break;
				case 'f':
					getParmString(&argc, &argv, parmString, 19);
					pFooterLength = atof(parmString);
					break;
				case 'g':
					getParmString(&argc, &argv, parmString, 19);
					pGapLength = atof(parmString);
					break;
				case 'h':
					getParmString(&argc, &argv, parmString, 19);
					pHeaderLength = atof(parmString);
					break;
				case 'k':
					getParmString(&argc, &argv, parmString, 19);
					sampleRate = atoi(parmString);
					break;
				case 'm':
					splitMultiLoad = true;
					break;
				case 's':
					getParmString(&argc, &argv, parmString, 19);
					startLoad = atoi(parmString);
					break;
                case 'v':
					getParmString(&argc, &argv, parmString, 19);
                    verbosity = atoi(parmString);
                    break;
				case 'w':
					getParmString(&argc, &argv, wavName, 255);
					break;
				default:
					cerr << "Invalid option " << (char)c << "\n";
					return 1;
					break;
			}

	if (verbosity)
	{
		cout << "AstroWav v1.3\n";
		cout << "Released Jul 27, 2002\n";
		cout << "Written by: Bob Colbert\n\n";
		cout.flush();
	}

    try
    {
        unsigned char *inLoad, *validLoad;

        inFile.open(argv[0],"r",verbosity);
		inFile.pHeaderLength = pHeaderLength;
		inFile.pFooterLength = pFooterLength;
		inFile.pGapLength = pGapLength;
        inFile.processTag("OPEN");
        inLoad = inFile.loadQue.front().dataBuffer;

        validFile.open("chickf2.bin","r",verbosity);
		validFile.pHeaderLength = pHeaderLength;
		validFile.pFooterLength = pFooterLength;
		validFile.pGapLength = pGapLength;
        validFile.processTag("OPEN");
        validLoad = validFile.loadQue.front().dataBuffer;

        for (int i = 0; i < 0x2000; i++)
        {
            if (inLoad[i] != validLoad[i])
            {
                printf("%4.4x\n", i);
            }
        }
    }

    catch(Error::endOfFile)
    {
        cerr << "Unexpected end of input reached\n";
        return 1;
    }

	catch(Error::badOutputFileName)
	{
		cerr << "Invalid output .wav filename\n";
		return 1;
	}

	return 0;
}
