#include <cstdlib>
#include <string>
#include <fstream>
#include <iostream>
#include <math.h>
#include "wordVal.h"

#define BLOCKSIZE 200000
//8192

#define UNDEFINED_MODE 0
#define INPUT_MODE 1
#define OUTPUT_MODE 2
#define WAV_HDR_FORMAT           0x14
#define WAV_HDR_CHANNELS         0x16
#define WAV_HDR_SAMPLE_RATE      0x18
#define WAV_HDR_BYTES_PER_SEC    0x1c
#define WAV_HDR_BYTES_PER_SAMPLE 0x20
#define WAV_HDR_BITS_PER_SAMPLE  0x22
#define WAV_HDR_LENGTH           0x28

#ifndef WAV_FILE_DEFINED

#define WAV_FILE_DEFINED

using namespace std;

class WavFile {
    friend class WavTimer;
public:
	WavFile();
	~WavFile();

	int    format, channels, bytesPerSample, bitsPerSample;
	long   sampleRate, bytesPerSecond;
	double length, time;

    int    openInput(char *filename);
	int    openOutput(char *pFileName, long pSampleRate = 44100, int pBitsPerSample = 8, int pChannels = 1);
    void   closeOutput();
    void   describeWav(void);
    bool   eof(void);
    double tellp(void);
    int    getSample(void);
    void   putSilence(double silenceLength);
	void   endOfFile();
protected:
    int           verbosity, mode, volume;
	std::ifstream      ifsWav;
	std::string   WavFileName;
    double        fileSize;
	long          blockNumber, bufferPosition;
	unsigned char buffer[BLOCKSIZE], byte, word[2], longWord[4];
	double        filePosition;

    void         validateMode(void);
	double       wordVal(double);
	double       signedWordVal(double);
	double       longWordVal(double);
	unsigned int getByte(double);
    void         getWord(double byteFilePosition, unsigned char *word);
	size_t       getBlock(long);
    void         putSample(int pAmplitude);
	unsigned char wavHeader[44];

private:
	std::ofstream      ofsWav;
};

class WavTimer {

public:
    WavTimer(WavFile *pWavFile);
    void reset();
    void set(streampos pFileByte);
    streampos getStartByte();
    double currentTime();

private:
    WavFile *wavFile;
    streampos startByte;
};

#endif