#ifdef _AFXDLL
    #include "stdafx.h"
#endif
#include "wordVal.h"
#include <iostream>
#include <cstdlib>
using namespace std;

double WordVal::wordVal(unsigned char *word)
{
	double temp;

    temp = (double)word[1];
	temp = (temp * 256) + (double)word[0];
	return temp;
}

double WordVal::signedWordVal(unsigned char *word)
{
    double temp;

    temp = wordVal(word);
	if (temp >= 0x8000)
	    temp = (0x10000 - temp) * -1;
	return temp;
}

double WordVal::longWordVal(unsigned char *longWord)
{
	double temp;

    temp = wordVal(longWord);
    temp += 0x10000 * wordVal(longWord+2);
	return temp;
}

double WordVal::signedLongWordVal(unsigned char *longWord)
{
	double temp;

    temp = wordVal(longWord);
    temp += 0x10000 * wordVal(longWord+2);
	if (temp >= 0x80000000)
	    temp = (0x100000000 - temp) * -1;
	return temp;
}

void WordVal::decToWord(long inVal, unsigned char *outWord)
{
    // Creates word in format LH
    unsigned char temp;
    long newWord;
    
    newWord = (inVal % 0x10000);

    if ((unsigned long)newWord >= (unsigned long)0x100)
        temp = (unsigned char)(((unsigned long)newWord - ((unsigned long)newWord%(unsigned long)0x100))/(unsigned long)0x100);
    else
        temp = 0;

    outWord[1] = temp;
    outWord[0] = newWord - (temp * 0x100);
}

void WordVal::decToSignedWord(double inVal, unsigned char *outLongWord)
{
    // Creates word in format LH
    long newWord;

    if (inVal < -(signed long)0x8000 || inVal > (signed long)0x7fff)
    {
        cerr << "Invalid signed word value\n";
        exit(1);
    }

    if (inVal < 0)
         newWord = (signed long)0x10000 + (signed long)inVal;

    decToWord((unsigned long)inVal, outLongWord);
}

void WordVal::decToLongWord(long inVal, unsigned char *outLongWord)
{
    // Creates long word in format L0H0L1H1 where L1H1 is the most significant word
    long newLongWord;

    newLongWord = (long)(inVal % 0x100000000L);

    decToWord(inVal % 0x10000, outLongWord);
    decToWord((unsigned long)inVal / (unsigned long)0x10000, outLongWord + 2);
}

void WordVal::decToSignedLongWord(double inVal, unsigned char *outLongWord)
{
    // Creates long word in format L0H0L1H1 where L1H1 is the most significant word
    long newLongWord;

    if (inVal < (signed long)0x80000000L || inVal > 0x7fffffff)
    {
        cerr << "Invalid signed long word value\n";
        exit(1);
    }

    if (inVal < 0)
         newLongWord = (signed long)0x100000000 + (signed long)inVal;

    decToLongWord((unsigned long)inVal, outLongWord);
}

void WordVal::putWord(unsigned char *buffer, int bufferPosition, long inVal)
{
    unsigned char wordVal[2];

    WordVal::decToWord(inVal, wordVal);
    buffer[bufferPosition] = wordVal[0];
    buffer[bufferPosition+1] = wordVal[1];
}

void WordVal::putLongWord(unsigned char *buffer, int bufferPosition, long inVal)
{
    unsigned char longWordVal[4];

    WordVal::decToLongWord(inVal, longWordVal);
    buffer[bufferPosition] = longWordVal[0];
    buffer[bufferPosition+1] = longWordVal[1];
    buffer[bufferPosition+2] = longWordVal[2];
    buffer[bufferPosition+3] = longWordVal[3];
}

void WordVal::strToBuffer(unsigned char *buffer, int bufferPosition, char *str)
{
    int i = 0;
    while (str[i] != 0)
    {
        buffer[bufferPosition+i] = str[i];
        i++;
    }
}

unsigned char WordVal::revByte(char inByte)
{
    char outByte = 0;
    int i;
    for (i=0; i<8; i++)
    {
        outByte <<= 1;
        if (inByte & 0x01)
        {
            outByte += 1;
        }
        inByte >>= 1;
    }
    return outByte;
}

